import json

import click
from cspark.sdk import Client, SparkError
from InquirerPy import inquirer
from rich.console import Console

from .._utils import get_active_profile
from ._api import AliasedGroup, header_option, json_parse, parse_pairs


@click.group(
    name='folders',
    help='Manage Spark folders',
    options_metavar='',
    cls=AliasedGroup,
    aliases={'ls': 'list', 'del': 'delete'},
)
def folders_cmd():
    pass


@folders_cmd.command(name='list', help='List available folders')
@click.option('--name-only', is_flag=True, default=False, help='Show folder names only')
@click.option('-d', '--data', type=str, metavar='<JSON>', help='Pagination data to send with the request')
@header_option()
def list_folders(name_only: bool, data: str, headers: list[str]) -> None:
    profile = get_active_profile()
    console = Console()

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))
        with client.folders as f:
            response = f.find(**json_parse(data)).data

        output = response['data']  # type: ignore
        if name_only:
            output = [item['name'] for item in output]  # type: ignore

        click.echo(json.dumps(output, indent=2))
    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc


@folders_cmd.command(name='delete', help='Delete a folder (and all its services)')
@click.argument('names', type=str, metavar='<FOLDER NAMES>', nargs=-1)
@header_option()
def delete_services(names: list[str], headers: list[str]) -> None:
    profile = get_active_profile()
    console = Console()

    if not names:
        raise click.BadParameter('at least one folder name must be provided.')

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))
        with client.folders as f:
            for folder in names:
                confirmed = inquirer.confirm(  # type: ignore
                    message=f'Are you sure you want to delete "{folder}" and all its services?', default=False
                ).execute()
                if not confirmed:
                    click.echo(f'folder skipped: {folder}')
                    continue

                response = f.delete(folder).data
                if response.get('status', '').lower() == 'success':  # type: ignore
                    click.echo(f'folder deleted: {folder}')

    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc



from sys import version_info as v

__all__ = ['__version__', '__title__', 'sdk_version', 'about']

__version__ = sdk_version = '0.2.4-rc.1'

__title__ = 'Coherent Spark Python SDK'

sdk_logger = f'CSPARK v{sdk_version}'

platform_info = f'Python/{v.major}.{v.minor}.{v.micro}'

about = f'Coherent Spark SDK v{sdk_version} ({platform_info})'

sdk_ua_header = f'agent=spark-python-sdk/{sdk_version}; env={platform_info}'



import click

from ._auth import auth_cmd
from ._config import config_cmd, init_cmd
from ._folders import folders_cmd
from ._services import services_cmd
from ._versions import versions_cmd


def register_commands(cli: click.Group) -> None:
    cli.add_command(init_cmd)
    cli.add_command(config_cmd)
    cli.add_command(auth_cmd)
    cli.add_command(versions_cmd)
    cli.add_command(services_cmd)
    cli.add_command(folders_cmd)



import logging
from dataclasses import dataclass
from typing import Any, Mapping, Union

from ._constants import DEFAULT_LOGGER_DATEFMT
from ._version import sdk_logger

__all__ = ['get_logger', 'LoggerOptions']


@dataclass
class LoggerOptions:
    context: str = sdk_logger
    disabled: bool = False
    level: int = logging.DEBUG
    datefmt: str = DEFAULT_LOGGER_DATEFMT
    colorful: bool = True
    timestamp: bool = True

    @staticmethod
    def when(options: Union[bool, Mapping[str, Any], 'LoggerOptions']) -> 'LoggerOptions':
        if isinstance(options, bool):
            return LoggerOptions(disabled=not options)
        if isinstance(options, Mapping):
            return LoggerOptions(**options)
        return options


class ColorfulFormatter(logging.Formatter):
    def __init__(self, *, datefmt: str, colorful: bool = True, timestamp: bool = True) -> None:
        super().__init__(datefmt=datefmt)
        self.colorful = colorful
        self.timestamp = timestamp

    def format(self, record: logging.LogRecord) -> str:
        timestamp = '%(asctime)s' if self.timestamp else ''
        if self.colorful:
            colored_heading = '\x1b[38;5;3m[%(name)s]\x1b[39m'  # orange
            colored_timestamp = f'\x1b[34m{timestamp}\x1b[39m'  # blue
            colored_level_and_msg = f'{self.color_by_level(record.levelno, "%(levelname)s - %(message)s")}'
            formatted_msg = f'{colored_heading} {colored_timestamp} {colored_level_and_msg}'
        else:
            formatted_msg = f'[%(name)s] {timestamp} %(levelname)s - %(message)s'

        record.levelname = record.levelname.rjust(8, ' ')  # right align levelname
        formatter = logging.Formatter(formatted_msg, datefmt=self.datefmt)
        return formatter.format(record)

    def color_by_level(self, level: int, msg: str) -> str:
        if level == logging.DEBUG:
            return f'\x1b[95m{msg}\x1b[39m'  # magenta
        elif level == logging.WARNING:
            return f'\x1b[93m{msg}\x1b[39m'  # yellow
        elif level == logging.ERROR:
            return f'\x1b[31m{msg}\x1b[39m'  # red
        elif level == logging.INFO:
            return f'\x1b[32m{msg}\x1b[39m'  # green
        elif level == logging.CRITICAL:
            return f'\x1b[1m{msg}\x1b[0m'  # bold
        else:
            return f'\x1b[32m{msg}\x1b[39m'  # green


def get_logger(
    context: str = sdk_logger,
    disabled: bool = False,
    level: int = logging.DEBUG,
    datefmt: str = DEFAULT_LOGGER_DATEFMT,
    colorful: bool = True,
    timestamp: bool = True,
) -> logging.Logger:
    logging.getLogger('httpx').disabled = True  # disable existing loggers

    logger = logging.getLogger(context)
    logger.setLevel(disabled and logging.CRITICAL or level)

    if not logger.handlers:
        formatter = ColorfulFormatter(datefmt=datefmt, colorful=colorful, timestamp=timestamp)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger



from cspark.sdk import ApiResource

from .._version import about

__all__ = ['HybridResource']


class HybridResource(ApiResource):
    @property
    def default_headers(self):
        return {**super().default_headers, 'User-Agent': about}



from ._base import *
from ._services import *

__all__ = ['Version', 'Health', 'Status']


class Version(HybridResource):
    def get(self):
        return self.request(f'{self.config.base_url.value}/version')


class Health(HybridResource):
    def check(self):
        return self.request(f'{self.config.base_url.value}/healthcheck')


class Status(HybridResource):
    def get(self):
        return self.request(f'{self.config.base_url.value}/status')



from cspark.sdk import sdk_version

__all__ = ['__version__', '__title__', 'about']

__version__ = '0.1.0'

__title__ = 'Coherent Hybrid Runner Python SDK'

about = f'Coherent Hybrid Runner SDK {__version__} (sdk v{sdk_version})'



from json import dumps
from os import getenv
from typing import Optional
from urllib.parse import urlparse

from cspark.sdk import BaseUrl, SparkError
from cspark.sdk import Config as BaseConfig
from httpx import Client as HttpClient

from ._constants import *

__all__ = ['Config', 'RunnerUrl']


class Config(BaseConfig):
    def copy_with(
        self,
        *,
        base_url: Optional[str] = None,
        tenant: Optional[str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[float] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        http_client: Optional[HttpClient] = None,
        **kwargs,  # noqa: ARG002
    ) -> 'Config':
        """Overrides parent's copyWith as `env` and `oauth` are not applicable."""
        url = (
            base_url.copy_with(tenant=tenant)
            if isinstance(base_url, BaseUrl)
            else self.base_url.copy_with(url=base_url, tenant=tenant)
        )
        return Config(
            base_url=url,
            api_key=api_key or self._auth.api_key,
            token=token or self._auth.token,
            timeout=timeout or self._timeout,
            max_retries=max_retries or self._max_retries,
            retry_interval=retry_interval or self._retry_interval,
            http_client=http_client or self.http_client,
        )


class RunnerUrl(BaseUrl):
    def __init__(self, url: Optional[str] = None, tenant: str = ''):
        super().__init__(url or getenv(ENV_VARS.RUNNER_URL) or DEFAULT_RUNNER_URL, tenant)

    def copy_with(self, *, url: Optional[str] = None, tenant: Optional[str] = None, **kwargs) -> 'RunnerUrl':  # noqa: ARG002
        """Overrides parent's copyWith as `env` is not applicable to RunnerUrl."""
        return RunnerUrl.of(url=url or self.value, tenant=tenant or self.tenant)

    @staticmethod
    def of(*, url: Optional[str] = None, tenant: Optional[str] = None, **kwargs) -> 'RunnerUrl':  # noqa: ARG004
        error_msg = 'cannot build base URL from invalid parameters'
        if url:
            parsed_url = urlparse(url.rstrip('/'))  # type: ignore
            paths = str(parsed_url.path).split('/')
            maybe_tenant = paths[1] if len(paths) > 1 else tenant

            if maybe_tenant:
                return RunnerUrl(url=f'{parsed_url.scheme}://{parsed_url.netloc}', tenant=str(maybe_tenant))
            else:
                error_msg = 'tenant name is required'
        elif tenant:
            return RunnerUrl(url='', tenant=str(tenant).strip().lower())

        raise SparkError.sdk(message=error_msg, cause=dumps({'url': url, 'tenant': tenant}))

    @staticmethod
    def no_tenant(url: str = getenv(ENV_VARS.RUNNER_URL) or DEFAULT_RUNNER_URL) -> 'RunnerUrl':
        """Builds a base URL from the given URL without a tenant (only valid for Version/Health API Resources)"""
        return RunnerUrl(url, tenant='')



import re
from abc import ABC, abstractmethod
from json import loads
from typing import List, Optional, Union

from ._errors import SparkError
from ._utils import StringUtils, is_positive_int

__all__ = ['Validators']


class BaseValidator(ABC):
    _errors: List[SparkError] = []

    @property
    def errors(self):
        return self._errors

    @abstractmethod
    def validate(self, *args, **kwargs):
        raise NotImplementedError

    def is_valid(self, *args, **kwargs) -> bool:
        try:
            self.validate(*args, **kwargs)
            return True
        except SparkError as error:
            self._errors.append(error)
            return False

    def reset(self):
        self._errors.clear()


class EmptyStringValidator(BaseValidator):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(EmptyStringValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Optional[str], message: Optional[str] = None):
        if StringUtils.is_empty(value):
            raise SparkError.sdk(message or 'must be non-empty string value', cause=value)


class PositiveNumberValidator(BaseValidator):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(PositiveNumberValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Union[None, int, float]):
        if not is_positive_int(value):
            raise SparkError.sdk('must be a positive number', value)


class BaseUrlValidator(BaseValidator):
    _wildcard = re.compile(r'^https://(?:[^./]+\.)+coherent\.global(?:/[^/?#]+)*(?:[?#].*)?$', re.IGNORECASE)

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(BaseUrlValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Optional[str]):
        if StringUtils.is_empty(value):
            raise SparkError.sdk('base URL is required', value)

        if not self._wildcard.match(str(value).rstrip('/')):
            raise SparkError.sdk('must be a Spark base URL <*.coherent.global>', value)


class TransformValidator(BaseValidator):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(TransformValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: str):
        if not value:
            raise SparkError.sdk('must be a properly defined JSON object', value)

        try:
            json_data = loads(value)
        except Exception as cause:
            raise SparkError.sdk(f'<{value}> must be a valid JSON object', cause) from cause

        if not json_data.get('transform_type'):
            raise SparkError.sdk("JSON object must have a 'transform_type' property", value)

        target_version = json_data.get('target_api_version')
        if not target_version or target_version not in ['v3', 'v4']:
            raise SparkError.sdk("'target_api_version' must be either 'v3' or 'v4'", value)


class Validators:
    empty_str = EmptyStringValidator
    positive_num = PositiveNumberValidator
    base_url = BaseUrlValidator
    transform = TransformValidator



from datetime import datetime
from typing import Any, BinaryIO, Optional, Union

from .._config import Config
from .._constants import SPARK_SDK
from .._errors import SparkApiError, SparkError
from .._utils import DateUtils, get_uuid
from ._base import ApiResource, Uri

__all__ = ['Folders']


class Folders(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.value, 'version': 'api/v1'}

    @property
    def categories(self) -> 'Categories':
        return Categories(self.config)

    def find(
        self,
        name: Optional[str] = None,
        *,
        favorite: Optional[bool] = None,
        page: int = 1,
        size: int = 100,
        sort: str = '-updated',
        **params: Any,
    ):
        url = Uri.of(None, endpoint='product/list', **self._base_uri)
        search = [{'field': k, 'value': v} for k, v in {**params, 'name': name, 'isstarred': favorite}.items() if v]
        body = {'search': search, 'page': page, 'pageSize': size, 'sort': sort}
        body.update({'shouldFetchActiveServicesCount': True})

        return self.request(url, method='POST', body=body)

    def create(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        category: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        launch_date: Union[None, str, int, datetime] = None,
        status: Optional[str] = None,
        cover: Optional[BinaryIO] = None,
    ):
        url = Uri.of(None, endpoint='product/create', **self._base_uri)
        startdate, launchdate = DateUtils.parse(start_date, launch_date)
        form = {
            'Name': name,
            'Description': description or f'Created by {SPARK_SDK}',
            'Category': category or 'Other',
            'StartDate': startdate.isoformat(),
            'LaunchDate': launchdate.isoformat(),
            'Status': status or 'Design',
        }
        response = self.request(url, method='POST', form=form)

        if response.status == 200 and isinstance(response.data, dict) and response.data['status'] == 'Success':
            if cover:
                self.upload_cover(response.data['data']['folderId'], cover)
            return self.request(response.data['data']['get_product_url'])

        cause = SparkApiError.to_cause(request=response.raw_request, response=response.raw_response)
        if response.data['errorCode'] == 'PRODUCT_ALREADY_EXISTS':  # type: ignore
            error = SparkError.api(409, {'message': f'folder name <{name}> already exists', 'cause': cause})
        else:
            error = SparkError.api(response.status, {'message': f'failed to create folder <{name}>', 'cause': cause})
        self.logger.error(error.message)
        raise error

    def update(
        self,
        id: str,
        *,
        description: Optional[str] = None,
        category: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        launch_date: Union[None, str, int, datetime] = None,
        cover: Optional[BinaryIO] = None,
    ):
        url = Uri.of(None, endpoint=f'product/update/{id}', **self._base_uri)
        if cover:
            self.upload_cover(id, cover)

        body: dict[str, Any] = {'shouldTrackUserAction': True}
        if description:
            body['description'] = description
        if category:
            body['category'] = category
        if DateUtils.is_date(start_date):
            body['startDate'] = DateUtils.to_datetime(start_date).isoformat()  # type: ignore
        if DateUtils.is_date(launch_date):
            body['launchDate'] = DateUtils.to_datetime(launch_date).isoformat()  # type: ignore
        return self.request(url, method='POST', body=body)

    def delete(self, id: str):
        url = Uri.of(None, endpoint=f'product/delete/{id}', **self._base_uri)
        self.logger.warning('deleting folder will also delete all its services')

        return self.request(url, method='DELETE')

    def upload_cover(self, id: str, image: BinaryIO, filename: Optional[str] = None):
        url = Uri.of(None, endpoint='product/uploadcoverimage', **self._base_uri)
        files = {'coverImage': filename and (filename, image) or image}

        return self.request(url, method='POST', form={'id': id}, files=files)


class Categories(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.value, 'version': 'api/v1'}

    def list(self):
        return self.request(Uri.of(None, endpoint='lookup/getcategories', **self._base_uri))

    def save(self, name: str, *, key: Optional[str] = None, icon: Optional[str] = None):
        url = Uri.of(None, endpoint='lookup/savecategory', **self._base_uri)
        body = {'value': name, 'key': key or get_uuid(), 'icon': icon or 'other.svg'}
        response = self.request(url, method='POST', body=body)
        return response.copy_with(data=self.__extract_data(response.data))

    def delete(self, key: str):
        url = Uri.of(None, endpoint=f'lookup/deletecategory/{key}', **self._base_uri)
        response = self.request(url, method='DELETE')
        return response.copy_with(data=self.__extract_data(response.data))

    def __extract_data(self, data: Any):
        if not isinstance(data, dict):
            return data

        categories = data.get('data', {}).get('Metadata.ProductCategories', [])
        return [{k.lower(): v for k, v in c.items()} for c in categories]



from typing import Any, Mapping, Optional, Union

import cspark.wasm.resources as API
from cspark.sdk import BaseUrl, LoggerOptions
from httpx import Client as HttpClient

from ._config import Config, RunnerUrl

__all__ = ['Client']


class Client:
    """
    The main entry point for the Hybrid Runner client.

    Hybrid Runners offer a smaller subset of functionality compared to the SaaS API.
    Thus, extending 'cspark.sdk' to support the Hybrid Runner API is a good way
    to keep the codebase consistent and maintainable.
    """

    def __init__(
        self,
        *,
        base_url: Union[None, str, BaseUrl] = None,
        tenant: Optional[str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[float] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        http_client: Optional[HttpClient] = None,
        logger: Union[bool, Mapping[str, Any], LoggerOptions] = True,
    ) -> None:
        self.config = Config(
            base_url=base_url if isinstance(base_url, BaseUrl) else RunnerUrl.of(url=base_url, tenant=tenant),
            api_key=api_key,
            token=token,
            timeout=timeout,
            max_retries=max_retries,
            retry_interval=retry_interval,
            http_client=http_client,
            logger=logger,
        )

    @property
    def version(self) -> API.Version:
        return API.Version(self.config)

    @property
    def health(self) -> API.Health:
        return API.Health(self.config)

    @property
    def status(self) -> API.Status:
        return API.Status(self.config)

    @property
    def services(self) -> API.Services:
        return API.Services(self.config)

    @staticmethod
    def health_check(base_url: Optional[str] = None, token: str = 'open', **options):
        config = Config(base_url=RunnerUrl.no_tenant(base_url or ''), token=token, **options)
        with API.Health(config) as health:
            return health.check()

    @staticmethod
    def get_version(base_url: Optional[str] = None, token: str = 'open', **options):
        config = Config(base_url=RunnerUrl.no_tenant(base_url or ''), token=token, **options)
        with API.Version(config) as version:
            return version.get()

    @staticmethod
    def get_status(base_url: Optional[str] = None, token: str = 'open', **options):
        config = Config(base_url=RunnerUrl.no_tenant(base_url or ''), token=token, **options)
        with API.Status(config) as status:
            return status.get()



from __future__ import annotations

import math
import random
import re
import uuid
from datetime import datetime, timedelta
from typing import Any, List, Optional, Tuple, Union, cast

from ._constants import DEFAULT_RETRY_INTERVAL, RETRY_RANDOMIZATION_FACTOR


def is_int(value: Any | None) -> bool:
    if isinstance(value, int):
        return True
    try:
        int(str(value))
        return True
    except ValueError:
        return False


def is_positive_int(value: Any | None) -> bool:
    return is_int(value) and cast(int, value) > 0


def sanitize_uri(url: str, leading: bool = False) -> str:
    sanitized = re.sub(r'/{2,}', '/', url)
    sanitized = sanitized.rstrip('/')
    return sanitized if leading else sanitized.lstrip('/')


def find_value_from_dict(key_path: str, data: dict, default: Any = None) -> Any:
    """
    Walks through a nested dictionary-based dataset and returns the value associated
    with the key path.
    """
    try:
        keys = key_path.split('.')
    except AttributeError:
        return default
    if not keys:
        return default

    # safe operations
    value = data.copy()
    for k in keys:
        value = cast(dict, value.get(k))
    return value


def get_retry_timeout(retries: int, interval: float = DEFAULT_RETRY_INTERVAL) -> float:
    randomization = random.uniform(0, 1) * RETRY_RANDOMIZATION_FACTOR
    return math.pow(2, retries) * interval * randomization


def is_not_empty_list(value: Any | None) -> bool:
    return isinstance(value, list) and len(value) > 0


def get_uuid() -> str:
    return str(uuid.uuid4())


class StringUtils:
    @staticmethod
    def is_str(value: Any | None) -> bool:
        return isinstance(value, str)

    @staticmethod
    def is_empty(value: Any | None) -> bool:
        return value is None or (isinstance(value, str) and value.strip() == '')

    @staticmethod
    def is_not_empty(value: Any | None) -> bool:
        return isinstance(value, str) and value.strip() != ''

    @staticmethod
    def mask(value: str, start: int = 0, end: int = 4, char: str = '*') -> str:
        if not value or start < 0 or end < 0:
            return value
        return value[:start] + char * (len(value) - start - end) + value[-end:]

    @staticmethod
    def join(value: Union[None, str, List[str]] = None, sep: str = ',') -> Union[str, None]:
        return sep.join(value) if isinstance(value, list) else value


class DateUtils:
    @staticmethod
    def is_date(value: Optional[int | str | datetime]) -> bool:
        if isinstance(value, datetime):
            return True
        if isinstance(value, (str, int)):
            try:
                if isinstance(value, int):
                    datetime.fromtimestamp(value / 1000)  # Treat value as a Unix timestamp in milliseconds.
                else:
                    datetime.fromisoformat(value)  # Handles ISO strings like "YYYY-MM-DD"
                return True
            except ValueError:
                return False
        return False

    @staticmethod
    def parse(
        start: Optional[int | str | datetime] = None,
        end: Optional[int | str | datetime] = None,
        *,
        years: int = 10,
        months: int = 0,
        days: int = 0,
    ) -> Tuple[datetime, datetime]:
        start_date = DateUtils.to_datetime(start) if start else datetime.now()

        # Calculate the end date
        if end and DateUtils.is_after(end, start_date):
            end_date = DateUtils.to_datetime(end)
        else:
            # Handling months overflow manually
            year = start_date.year + years
            month = start_date.month + months
            day = start_date.day + days

            while month > 12:
                month -= 12
                year += 1

            end_date = start_date.replace(year=year, month=month) + timedelta(days=day)

        return start_date, end_date

    @staticmethod
    def is_before(date: Union[str, int, datetime], when: datetime) -> bool:
        return DateUtils.to_datetime(date) < when

    @staticmethod
    def is_after(date: Union[str, int, datetime], when: datetime) -> bool:
        return DateUtils.to_datetime(date) > when

    @staticmethod
    def to_datetime(value: Union[str, int, datetime]) -> datetime:
        if isinstance(value, datetime):
            return value
        if isinstance(value, int):
            return datetime.fromtimestamp(value / 1000)  # Unix timestamp in milliseconds
        if isinstance(value, str):
            try:
                return datetime.fromisoformat(value)
            except ValueError:
                return datetime.strptime(value, '%Y-%m-%d')



from __future__ import annotations

import re
import time
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Mapping, Optional, Union

from httpx import URL, Client, Headers, HTTPError, HTTPStatusError, Request, RequestError, Response

from .._config import Config
from .._errors import SparkApiError, SparkError
from .._logger import get_logger
from .._utils import StringUtils, get_retry_timeout, get_uuid, sanitize_uri
from .._version import about, sdk_ua_header

__all__ = ['ApiResource', 'UriParams', 'Uri', 'HttpResponse']


class ApiResource:
    def __init__(self, config: Config):
        self.config = config
        self.logger = get_logger(**config.logger.__dict__)
        self._client = config.http_client or Client(timeout=config.timeout / 1000)

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        # Ideally, let users handle the closing of the custom HTTP client.
        if self.config.http_client:
            return
        self.close()

    def close(self):
        if not self._client.is_closed:
            self._client.close()

    @property
    def default_headers(self):
        return {
            **self.config.extra_headers,
            'User-Agent': about,
            'x-spark-ua': sdk_ua_header,
            'x-request-id': get_uuid(),
            'x-tenant-name': self.config.base_url.tenant,
        }

    def request(
        self,
        url: Union[str, 'Uri', URL],
        *,
        method: str = 'GET',
        headers: Mapping[str, str] = {},
        params: Optional[Mapping[str, str]] = None,
        body: Optional[Any] = None,
        content: Optional[bytes] = None,
        form=None,
        files=None,
    ) -> 'HttpResponse':
        url = str(url)
        request = self._client.build_request(
            method,
            url,
            params=params,
            headers={**headers, **self.default_headers},
            data=form,
            json=body,
            content=content,
            files=files,
        )

        self.logger.debug(f'{method} {url}')
        return self.__fetch(request)

    def __fetch(self, request: Request, retries: int = 0) -> 'HttpResponse':
        request.headers.update(self.config.auth.as_header)

        response, status_code = None, 0
        err_msg = f'an error occurred while fetching <{request.url}>'

        try:
            response = self._client.send(request)
            response.raise_for_status()
        except RequestError as err:
            err_msg += f'; {err}'  # occurs while issuing a request; hence no response
            raise SparkError.sdk(err_msg, SparkApiError.no_response(request)) from err
        except HTTPStatusError as err:
            err_msg = str(err)
        except HTTPError as err:
            err_msg += f'; {err}'
        except Exception:
            pass  # possibly runtime error but should not interrupt this flow

        if not response:
            raise SparkError.sdk(err_msg, SparkApiError.no_response(request))

        status_code = response.status_code
        if status_code >= 400:
            if status_code == 401 and self.config.auth.type == 'oauth' and retries < self.config.max_retries:
                self.config.auth.oauth.retrieve_token(self.config)  # pyright: ignore[reportOptionalMemberAccess]
                return self.__fetch(request, retries + 1)

            if (status_code == 408 or status_code == 429) and retries < self.config.max_retries:
                self.logger.debug(f'retrying request due to status code {status_code}...')
                delay = get_retry_timeout(retries, self.config.retry_interval)
                time.sleep(delay)
                return self.__fetch(request, retries + 1)

            raise SparkError.api(
                status_code,
                {'message': f'failed to fetch <{request.url}>', 'cause': SparkApiError.to_cause(request, response)},
            )

        # otherwise, ok response
        ok_response = {
            'status': status_code,
            'data': None,
            'buffer': response.content,
            'headers': response.headers,
            'raw_request': request,
            'raw_response': response,
        }

        content_type = response.headers.get('content-type', '')
        if 'application/json' in content_type:
            try:
                ok_response['data'] = response.json()
            except Exception:
                ok_response['data'] = response.text
        return HttpResponse(**ok_response)


def download(
    url: str,
    *,
    method: str = 'GET',
    headers: Mapping[str, str] = {},
    params: Optional[Mapping[str, str]] = None,
    body=None,
    form=None,
    timeout: Optional[float] = 60,
):
    with Client() as client:
        request = client.build_request(
            method,
            url,
            params=params,
            headers=headers,
            data=form,
            json=body,
            timeout=timeout,
        )

        response = client.send(request)
        if response.status_code >= 400:
            raise SparkError.api(response.status_code, SparkApiError.to_cause(request, response))

        return HttpResponse(
            status=response.status_code,
            data=None,
            buffer=response.content,
            headers=response.headers,
            raw_request=request,
            raw_response=response,
        )


@dataclass
class HttpResponse:
    status: int
    data: Union[None, Any, str]
    buffer: bytes
    headers: Headers
    raw_request: Request
    raw_response: Response

    def copy_with(self, **kwargs) -> 'HttpResponse':
        return HttpResponse(
            status=kwargs.get('status', self.status),
            data=kwargs.get('data', self.data),
            buffer=kwargs.get('buffer', self.buffer),
            headers=kwargs.get('headers', self.headers),
            raw_request=kwargs.get('request', self.raw_request),
            raw_response=kwargs.get('response', self.raw_response),
        )


@dataclass(frozen=True)
class UriParams:
    folder: Optional[str] = None
    service: Optional[str] = None
    service_id: Optional[str] = None
    version: Optional[str] = None
    version_id: Optional[str] = None
    proxy: Optional[str] = None
    public: Optional[bool] = False

    @property
    def service_uri(self) -> str:
        """
        Returns the service URI locator's short format if available.
        folder/service[version?]
        """
        return Uri.encode(self, long=False)

    def pick(self, *args: str) -> 'UriParams':
        return UriParams(**{k: v for k, v in self.__dict__.items() if k in args})

    def omit(self, *args: str) -> 'UriParams':
        return UriParams(**{k: v for k, v in self.__dict__.items() if k not in args})

    def encode(self, long: bool = True) -> str:
        return Uri.encode(self, long=long)


class Uri:
    def __init__(self, url: URL):
        self._url = url

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f'<Uri: {self.value}>'

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Uri):
            return False
        return self.value == other.value

    @property
    def value(self) -> str:
        return str(self._url)

    @staticmethod
    def of(
        uri: Optional[UriParams] = None,
        *,
        base_url: str = '',
        version: str = 'api/v3',
        endpoint: str = '',
    ) -> 'Uri':
        uri = uri or UriParams()
        path = version
        if uri.public:
            path += '/public'

        if uri.version_id:
            path += f'/version/{uri.version_id}'
        elif uri.service_id:
            path += f'/service/{uri.service_id}'
        elif uri.folder and uri.service:
            path += f'/folders/{uri.folder}/services/{uri.service}'
        elif uri.proxy:
            path += f'/proxy/{sanitize_uri(uri.proxy)}'

        if endpoint and not uri.proxy:
            path += f'/{endpoint}'

        try:
            return Uri(URL(f'{base_url}/{path}'))
        except Exception as cause:
            raise SparkError.sdk('invalid URI params', uri) from cause

    @staticmethod
    def partial(
        resource: str,
        *,
        base_url: str = '',
        version: str = 'api/v3',
        endpoint: str = '',
    ) -> 'Uri':
        try:
            resource = sanitize_uri(resource)
            if version:
                resource = f'{version}/{resource}'
            if endpoint:
                resource += f'/{endpoint}'
            resource = f'{base_url}/{resource}'

            return Uri(URL(resource))
        except SparkError as error:
            raise SparkError.sdk(f'invalid service URI <{resource}>', error) from error
        except Exception as cause:
            raise SparkError(f'failed to build Spark endpoint from <{resource}>', cause) from cause

    @staticmethod
    def to_params(uri: Union[str, UriParams]) -> UriParams:
        return Uri.decode(uri) if isinstance(uri, str) else uri

    @staticmethod
    def decode(uri: str) -> UriParams:
        uri = re.sub('folders/', '', sanitize_uri(uri))
        uri = re.sub('services/', '', uri)
        match = re.match(r'^([^\/]+)\/([^[]+)(?:\[(.*?)\])?$', uri)
        if match:
            folder, service, version = match.groups()
            if folder == 'version':
                return UriParams(version_id=service)
            if folder == 'service':
                return UriParams(service_id=service)
            if folder == 'proxy':
                return UriParams(proxy=service)
            version = None if StringUtils.is_empty(version) else version
            return UriParams(folder=folder, service=service, version=version)
        return UriParams()

    @staticmethod
    def encode(uri: UriParams, long: bool = True) -> str:
        folder, service, version = uri.folder, uri.service, uri.version
        if uri.proxy:
            return f'proxy/{uri.proxy}'
        if uri.version_id:
            return f'version/{uri.version_id}'
        if uri.service_id:
            return f'service/{uri.service_id}'
        if folder and service:
            version = f'[{version}]' if version else ''
            if long:
                return f'folders/{folder}/services/{service}{version}'
            return f'{folder}/{service}{version}'
        return ''

    @staticmethod
    def validate(uri: Union[str, UriParams], message: Optional[str] = None) -> UriParams:
        uri_params = Uri.to_params(uri)
        if StringUtils.is_empty(uri_params.service_uri):
            folder, service = uri_params.folder, uri_params.service
            if folder and not service:
                msg = message or 'service name is missing'
            elif not folder and service:
                msg = message or 'folder name is missing'
            else:
                msg = message or 'service uri locator is required'
            msg += ' :: a uri needs to be of these formats: \n\t- "folder/service[?version]" \n\t-'
            msg += ' "service/service_id" \n\t- "version/version_id" \n\t- "proxy/custom-endpoint"\n'
            raise SparkError.sdk(msg, uri)
        return uri_params



from dataclasses import dataclass
from typing import Any, Union

from ._base import ApiResource, Uri

__all__ = ['OAuth2', 'AccessToken']


class OAuth2(ApiResource):
    def gen_access_token(self):
        url = Uri.of(base_url=self.config.base_url.oauth2, version='protocol', endpoint='openid-connect/token')
        body = {**self.config.auth.oauth.to_dict(), 'grant_type': self.config.auth.oauth.flow}  # type: ignore

        return self.request(url, method='POST', form=body)

    def get_access_token(self) -> 'AccessToken':
        """Retrieves the access token from the OAuth2 server."""
        response = self.gen_access_token()
        return AccessToken.from_dict(response.data)


@dataclass(frozen=True)
class AccessToken:
    access_token: str
    expires_in: int
    refresh_expires_in: int
    token_type: str
    not_before_policy: int
    scope: str

    @staticmethod
    def from_dict(data: Union[None, Any, str]) -> 'AccessToken':
        if not isinstance(data, dict):
            data = {}

        return AccessToken(
            access_token=str(data.get('access_token', '')),
            expires_in=int(data.get('expires_in', 0)),
            refresh_expires_in=int(data.get('refresh_expires_in', 0)),
            token_type=str(data.get('token_type', '')),
            not_before_policy=int(data.get('not-before-policy', 0)),
            scope=str(data.get('scope', '')),
        )



import json
import math
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union, cast

from .._config import Config
from .._constants import SPARK_SDK
from .._errors import SparkError
from .._utils import StringUtils, get_uuid, is_not_empty_list
from ._base import ApiResource, Uri, UriParams

__all__ = ['Batches', 'Pipeline', 'BatchChunk', 'ChunkData', 'create_chunks']


@dataclass
class ChunkData:
    inputs: List[Any]
    parameters: Optional[Dict[str, Any]] = None
    summary: Optional[Dict[str, Any]] = None

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'ChunkData':
        """Creates a chunk data object from dictionary-based data."""
        return ChunkData(data.get('inputs', []), data.get('parameters', {}), data.get('summary'))

    @staticmethod
    def from_str(data: Union[str, bytes]) -> 'ChunkData':
        """Creates a chunk data object from string-based data."""
        try:
            return ChunkData.from_dict(json.loads(data))
        except json.JSONDecodeError as err:
            raise SparkError.sdk('failed to parse string/bytes data as JSON', cause=err) from err
        except Exception as exc:
            raise SparkError.sdk(f'cannot create chunk data from {data}', cause=exc) from exc


@dataclass
class BatchChunk:
    id: str
    data: ChunkData
    size: Optional[int] = None

    def to_dict(self):
        return {'id': self.id, 'data': self.data.to_dict(), 'size': self.size or len(self.data.inputs)}

    @staticmethod
    def from_dict(data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> List['BatchChunk']:
        """Creates a list of batch chunks from dictionary-based data."""
        if isinstance(data, dict):
            data = [data]

        chunks = []
        for chunk in data:
            if not isinstance(chunk, dict):
                continue

            id = chunk.get('id', get_uuid())
            chunk_data = ChunkData.from_dict(chunk.get('data', {}))
            size = chunk.get('size', len(chunk_data.inputs))

            chunks.append(BatchChunk(id, data=chunk_data, size=size))
        return chunks

    @staticmethod
    def from_str(data: Union[str, bytes]) -> List['BatchChunk']:
        """Creates a list of batch chunks from string-based data."""
        try:
            json_data = json.loads(data)
            chunks = json_data.pop('chunks', []) if 'chunks' in json_data else json_data
            return BatchChunk.from_dict(chunks)
        except json.JSONDecodeError as err:
            raise SparkError.sdk('failed to parse string/bytes data as JSON', cause=err) from err
        except Exception as exc:
            raise SparkError.sdk(f'cannot create batch chunks from {data}', cause=exc) from exc


class Batches(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.full, 'version': 'api/v4'}

    def describe(self):
        """Describes the batch pipeline status across the tenant."""
        url = Uri.of(None, endpoint=f'batch/status', **self._base_uri)
        return self.request(url, method='GET')

    def create(
        self,
        uri: Union[str, UriParams],
        *,
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        unique_record_key: Union[None, str, List[str]] = None,
        # Experimental parameters (likely to change/be deprecated in future releases)
        min_runners: Optional[int] = None,
        max_runners: Optional[int] = None,
        chunks_per_vm: Optional[int] = None,
        runners_per_vm: Optional[int] = None,
        max_input_size: Optional[float] = None,
        max_output_size: Optional[float] = None,
        accuracy: Optional[float] = None,
    ):
        uri = Uri.validate(uri)
        url = Uri.of(base_url=self.config.base_url.full, version='api/v4', endpoint='batch')
        service_uri = uri.pick('folder', 'service', 'version').encode(long=False)
        body = {
            'service': uri.service_id or service_uri or None,
            'version_id': uri.version_id,
            'version_by_timestamp': active_since,
            'subservice': StringUtils.join(subservices),
            'output': StringUtils.join(selected_outputs),
            'call_purpose': call_purpose or 'Async Batch Execution',
            'source_system': source_system or SPARK_SDK,
            'correlation_id': correlation_id,
            'unique_record_key': StringUtils.join(unique_record_key),
            # experimental pipeline options
            'initial_workers': min_runners,
            'max_workers': max_runners,
            'chunks_per_request': chunks_per_vm,
            'runner_thread_count': runners_per_vm,
            'max_input_size': max_input_size,
            'max_output_size': max_output_size,
            'acceptable_error_percentage': math.ceil((1 - min(accuracy or 1, 1.0)) * 100),
        }

        return self.request(url, method='POST', body={k: v for k, v in body.items() if v is not None})

    def of(self, batch_id: str) -> 'Pipeline':
        return Pipeline(batch_id, self.config)


class Pipeline(ApiResource):
    _state: str = 'open'
    _chunks: Dict[str, int] = dict()

    def __init__(self, batch_id: str, config: Config):
        super().__init__(config)
        self._id = batch_id
        self._base_uri = {'base_url': self.config.base_url.full, 'version': 'api/v4'}

        if StringUtils.is_empty(batch_id):
            error = SparkError.sdk('batch pipeline id is required to proceed', batch_id)
            self.logger.error(error.message)
            raise error

    @property
    def batch_id(self):
        return self._id

    @property
    def state(self):
        return self._state

    @property
    def stats(self):
        return {'chunks': len(self._chunks), 'records': sum(self._chunks.values())}

    @property
    def is_disposed(self) -> bool:
        """
        Determines if the batch pipeline is no longer active.

        This is for the internal use of the SDK and does not validate the actual state
        of the batch pipeline. Use `getStatus()` to get the actual state of the batch
        pipeline.
        """
        return self._state == 'closed' or self._state == 'cancelled'

    def get_info(self):
        """Retrieves the batch pipeline info."""
        url = Uri.of(None, endpoint=f'batch/{self._id}', **self._base_uri)
        return self.request(url, method='GET')

    def get_status(self):
        """Retrieves the batch pipeline status."""
        url = Uri.of(None, endpoint=f'batch/{self._id}/status', **self._base_uri)
        return self.request(url, method='GET')

    def push(
        self,
        *,
        chunks: Optional[List[BatchChunk]] = None,
        data: Optional[ChunkData] = None,
        inputs: Optional[List[Any]] = None,
        raw: Union[None, str, bytes] = None,
        if_chunk_id_duplicated: str = 'replace',  # 'ignore' | 'replace' | 'throw'
    ):
        self.__assert_state(['closed', 'cancelled'])

        url = Uri.of(None, endpoint=f'batch/{self._id}/chunks', **self._base_uri)
        body = self.__build_chunk_data(chunks, data, inputs, raw, if_chunk_id_duplicated)

        response = self.request(url, method='POST', body=body)
        total = response.data['record_submitted'] if isinstance(response.data, dict) else 0
        self.logger.info(f'pushed {total} records to batch pipeline <{self._id}>')

        return response

    def pull(self, max_chunks: int = 100):
        self.__assert_state(['cancelled'])

        endpoint = f'batch/{self._id}/chunkresults?max={max(1, max_chunks)}'

        response = self.request(Uri.of(None, endpoint=endpoint, **self._base_uri))
        total = response.data['status']['records_available'] if isinstance(response.data, dict) else 0
        self.logger.info(f'{total} available records from batch pipeline <{self._id}>')

        return response

    def dispose(self, state: str = 'closed'):
        """Disposes the batch pipeline."""
        self.__assert_state(['closed', 'cancelled'])

        url = Uri.of(None, endpoint=f'batch/{self._id}', **self._base_uri)

        response = self.request(url, method='PATCH', body={'batch_status': state})
        self.logger.info(f'batch pipeline <{self._id}> has been {state}')
        self._state = state

        return response

    def cancel(self):
        return self.dispose(state='cancelled')

    def __assert_state(self, states: List[str], throwable: bool = True) -> bool:
        if self._state in states:
            error = SparkError.sdk(f'batch pipeline <{self._id}> is already {self._state}')
            self.logger.error(error.message)
            if throwable:
                raise error
            return False
        return True

    def __build_chunk_data(
        self,
        chunks: Optional[List[BatchChunk]] = None,
        data: Optional[ChunkData] = None,
        inputs: Optional[List[Any]] = None,
        raw: Union[None, str, bytes] = None,
        if_duplicated: str = 'replace',
    ):
        try:
            if raw is not None and StringUtils.is_not_empty(raw):
                chunks = BatchChunk.from_str(raw)  # takes precedence over other entries

            if is_not_empty_list(chunks):
                return {'chunks': self.__assess_chunks(cast(List[BatchChunk], chunks), if_duplicated)}
            if isinstance(data, ChunkData) and is_not_empty_list(data.inputs):
                return {'chunks': self.__assess_chunks([BatchChunk(id=get_uuid(), data=data)], if_duplicated)}
            if is_not_empty_list(inputs):
                data = ChunkData(inputs=cast(List[Any], inputs), parameters={})
                return {'chunks': self.__assess_chunks([BatchChunk(id=get_uuid(), data=data)], if_duplicated)}

            cause = {'chunks': chunks, 'data': data, 'inputs': inputs, 'raw': raw}
            raise SparkError.sdk(
                message=f'wrong data params were provided for this pipeline <{self._id}>.\n'
                'Expecting either "raw=str/bytes", "chunks=List[BatchChunk]", "data=ChunkData" or "inputs=List[Any]"',
                cause=json.dumps(cause),
            )
        except SparkError as error:
            self.logger.error(error.message)
            raise
        except Exception as cause:
            raise SparkError.sdk('failed to build push data for batch pipeline', cause) from cause

    def __assess_chunks(self, chunks: List['BatchChunk'], is_duplicated: str):
        assessed = []
        for chunk in chunks:
            id = get_uuid() if StringUtils.is_empty(chunk.id) else chunk.id
            if chunk.id in self._chunks:
                if is_duplicated == 'ignore':
                    self.logger.warn(
                        f'chunk id <{id}> appears to be duplicated for this pipeline <{self._id}> '
                        f'and may cause unexpected behavior. You should consider using a different id.'
                    )
                    continue
                if is_duplicated == 'throw':
                    raise SparkError.sdk(
                        message=f'chunk id <{id}> is duplicated for batch pipeline <{self._id}>',
                        cause=chunk,
                    )
                if is_duplicated == 'replace':
                    self._chunks.pop(chunk.id)  # remove the old chunk id
                    chunk.id = get_uuid()  # add new id for the chunk
                    self.logger.info(
                        f'chunk id <{id}> is duplicated for this pipeline <{self._id}> '
                        f'and has been replaced with <{chunk.id}>'
                    )
            self._chunks[chunk.id] = chunk.size or len(chunk.data.inputs) - 1
            assessed.append(chunk.to_dict())
        return assessed


def create_chunks(
    dataset: List[Any],  # input dataset in json array format
    *,
    headers: Optional[List[str]] = None,  # headers for the input dataset
    chunk_size: int = 200,
    parameters: Optional[Dict[str, Any]] = None,
    summary: Optional[Dict[str, Any]] = None,
) -> List[BatchChunk]:
    """Creates a list of batch chunks from a JSON array dataset."""
    length = len(dataset)
    chunk_size = max(1, chunk_size)
    batch_size = math.ceil(length / chunk_size)

    headers = headers or (length > 0 and cast(List[str], dataset.pop(0))) or []
    if not headers:
        raise SparkError.sdk('missing headers for the input dataset', cause=dataset)

    chunks = []
    for i in range(batch_size):
        start = i * chunk_size
        end = min(start + chunk_size, length)
        inputs = [headers] + dataset[start:end]
        chunks.append(BatchChunk(get_uuid(), ChunkData(inputs, parameters or {}, summary), size=len(inputs) - 1))
    return chunks



import json

import click
from cspark.sdk import Client, Config, SparkError
from rich.console import Console

from .._utils import get_active_profile
from ._api import AliasedGroup, Services, header_option, json_parse, params_option, parse_pairs


@click.group(
    name='services',
    help='Manage Spark services',
    options_metavar='',
    cls=AliasedGroup,
    aliases={'ls': 'list', 'exec': 'execute', 'run': 'execute', 'find': 'search', 'del': 'delete'},
)
def services_cmd():
    pass


@services_cmd.command(name='list', help='List available services in a folder')
@click.argument('folder', type=str, metavar='<FOLDER NAME>')
@click.option('--name-only', is_flag=True, default=False, help='Show service names only')
@click.option('-d', '--data', type=str, metavar='<JSON>', help='Pagination data to send with the request')
@header_option()
def list_services(folder: str, name_only: bool, data: str, headers: list[str]) -> None:
    profile = get_active_profile()
    console = Console()

    try:
        config = Config(**profile.to_config())
        config.extra_headers.update(parse_pairs(headers))
        with Services(config) as s:
            response = s.get(folder, data).data

        output = response['data']  # type: ignore
        if name_only:
            output = [item['serviceName'] for item in output]  # type: ignore

        click.echo(json.dumps(output, indent=2))
    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc


@services_cmd.command(name='search', help='Search available services')
@click.option('-d', '--data', type=str, metavar='<JSON>', help='Pagination data to send with the request')
@header_option()
@click.option('--show-all', is_flag=True, default=False, help='Show the complete response data')
def search_services(data: str, headers: list[str], show_all: bool) -> None:
    profile = get_active_profile()
    console = Console()

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))
        with client.services as s:
            response = s.search(**json_parse(data))

        output = response.data
        if not show_all:
            output = output.get('response_data', {})  # type: ignore

        click.echo(json.dumps(output, indent=2))
    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc


@services_cmd.command(name='execute', help='Execute Spark services')
@click.argument('uri', type=str, metavar='<SERVICE URI>')
@click.option('-i', '--inputs', type=str, metavar='<JSON>', help='Input data to send with the request')
@params_option()
@header_option()
@click.option('--show-all', is_flag=True, default=False, help='Show the complete response data')
def execute_services(uri: str, inputs: str, params: list[str], headers: list[str], show_all: bool) -> None:
    profile = get_active_profile()
    console = Console()

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))
        with client.services as s:
            metadata = parse_pairs(params, infer_type=True)
            outputs = s.execute(uri, inputs=inputs, source_system='Coherent Spark CLI', **metadata).data

        if not show_all:
            outputs = outputs.get('outputs') or outputs.get('response_data', {}).get('outputs')  # type: ignore

        click.echo(json.dumps(outputs, indent=2))
    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc


@services_cmd.command(name='delete', help='Delete services in a folder (irreversible - use with caution)')
@click.option('-f', '--folder', type=str, metavar='<FOLDER NAME>', help='Name of the folder to delete services from')
@click.argument('names', type=str, metavar='<SERVICE NAMES>', nargs=-1)
@header_option()
def delete_services(folder: str, names: list[str], headers: list[str]) -> None:
    profile = get_active_profile()
    console = Console()

    if not names:
        raise click.BadParameter('at least one service name must be provided.')

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))

        with client.services as s:
            for service in names:
                response = s.delete(folder=folder, service=service).data
                if response.get('status', '').lower() == 'success':  # type: ignore
                    click.echo(f'service deleted: {service}')

    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc



import click
from cspark.sdk import sdk_version

from ._commands import register_commands

__version__ = '0.1.0-beta'
__title__ = f'Coherent Spark CLI {__version__} (sdk v{sdk_version})'
__description__ = 'Interact with Coherent Spark APIs from the command line'


@click.group(invoke_without_command=True)
@click.pass_context
@click.version_option(__version__, '-v', '--version', message=__title__)
def main(ctx: click.Context) -> None:
    if ctx.invoked_subcommand is None:
        click.echo(f'{__description__}.\n')
        click.echo(ctx.get_help())
        click.echo('\nFor more details, visit https://docs.coherent.global/.')


register_commands(main)



import json
from typing import Any, Callable, Mapping, Optional, Union

import click
from cspark.sdk import ApiResource, Uri


def header_option(**kwargs: Any) -> Callable:
    """Define a decorator that applies headers to a command"""

    def decorator(func: Union[Callable, click.Command]) -> Union[Callable, click.Command]:
        return click.option(
            '-H',
            '--header',
            'headers',
            type=str,
            multiple=True,  # allows multiple headers
            help='Headers to send with the request (e.g., "x-my-header: value")',
            metavar='<HEADER>',
            **kwargs,
        )(func)

    return decorator


def params_option(**kwargs: Any) -> Callable:
    """Define a decorator that applies parameters to a command"""

    def decorator(func: Union[Callable, click.Command]) -> Union[Callable, click.Command]:
        return click.option(
            '-p',
            '--param',
            'params',
            type=str,
            multiple=True,  # allows multiple params
            help='Additional parameters (or metadata) supported',
            metavar='<PARAM>',
            **kwargs,
        )(func)

    return decorator


def parse_pairs(pairs: list[str], *, sep: str = ':', infer_type: bool = False) -> dict:
    """Parse a list of key-value pairs into a dictionary"""

    def _infer_type(value: str):
        val = value.strip().lower()
        if val == 'true':
            return True
        elif val == 'false':
            return False
        try:
            return float(val) if '.' in val else int(val)
        except ValueError:
            return value  # default to string

    parsed = {}
    for pair in pairs:
        try:
            key, value = pair.split(sep, 1)
            parsed[key.strip()] = _infer_type(value.strip()) if infer_type else value.strip()
        except Exception:
            pass

    return parsed


def json_parse(data: str, fallback: Optional[Any] = None) -> dict:
    try:
        return json.loads(data)
    except Exception:
        return fallback or {}


class Services(ApiResource):
    def get(self, folder: str, data: Optional[str] = None):
        endpoint = f'product/{folder}/engines'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)
        body = {'page': 1, 'pageSize': 100, 'search': [], 'sort': 'name1'}
        return self.request(url, method='POST', body=body.update(json_parse(data)) if data else body)


class AliasedGroup(click.Group):
    """A click Group that allows aliases to be defined"""

    def __init__(self, *, name: str, aliases: Optional[Mapping[str, str]] = None, **attrs: Any) -> None:
        self.aliases = aliases or {}
        super().__init__(name=name, **attrs)

    def get_command(self, ctx: click.Context, cmd_name: str):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        if cmd_name in self.aliases:
            actual_cmd = self.aliases[cmd_name]
            return click.Group.get_command(self, ctx, actual_cmd)

        # NOTE: is this a good idea?
        # Alternative option: if we did not find an explicit alias we
        # allow automatic abbreviation of the command.  "execute" for
        # instance will match "exec".  We only allow that however if
        # there is only one command.
        # matches = [x for x in self.list_commands(ctx) if x.lower().startswith(cmd_name.lower())]

        # if not matches:
        #     return None
        # elif len(matches) == 1:
        #     return click.Group.get_command(self, ctx, matches[0])
        # ctx.fail(f"too many matches: {', '.join(sorted(matches))}")
        return None

    def resolve_command(self, ctx, args):
        # always return the command's name, not the alias
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args  # type: ignore



import json
import pathlib
from datetime import datetime, timedelta
from typing import Union

import click
from cspark.sdk import Config, OAuth, SparkApiError
from rich.console import Console

from .._utils import DATE_FORMAT, HOME_DIR, Profile, get_active_profile, load_profiles


@click.group(name='auth', help='Manage Spark authorization with OAuth2 CCG', options_metavar='')
def auth_cmd():
    pass


def _friendly_time(seconds: Union[int, float]) -> str:
    if seconds < 60:
        return f'{int(seconds)} second{"s" if seconds > 1 else ""}'
    elif seconds < 3600:
        mins = seconds // 60
        return f'{int(mins)} min{"s" if mins > 1 else ""}'
    elif seconds < 86400:
        hrs = seconds // 3600
        mins_left = (seconds % 3600) // 60
        timeago = f'{int(hrs)} hr{"s" if hrs != 1 else ""}'
        if mins_left > 0:
            timeago = f'{timeago} {int(mins_left)} min{"s" if mins_left > 1 else ""}'
        return timeago
    else:
        days = seconds // 86400
        hrs_left = (seconds % 86400) // 3600
        timeago = f'{int(days)} day{"s" if days != 1 else ""}'
        if hrs_left > 0:
            timeago = f'{timeago} {int(hrs_left)} hr{"s" if hrs_left > 1 else ""}'
        return timeago


class AuthStatusCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='status',
            help='Display authentication status of the active profile',
            callback=self.status,
            params=[
                click.Option(
                    ['-a', '--all'],
                    is_flag=True,
                    default=False,
                    type=bool,
                    help='Display authentication status of all the profiles',
                ),
                click.Option(
                    ['-v', '--verbose'],
                    is_flag=True,
                    default=False,
                    type=bool,
                    help='Display additional details of the authentication status',
                ),
            ],
        )

    def status(self, all: bool, verbose: bool):
        profiles = load_profiles()

        if all:
            for profile in profiles:
                self._display_status(profile, verbose=verbose)
        else:
            self._display_status(get_active_profile(), verbose=verbose)

    def _display_status(self, profile: Profile, verbose: bool):
        path = HOME_DIR / f'{profile.name}_auth.json'
        if path.exists():
            with path.open('r') as file:
                auth = json.load(file)
                auth.pop('profile')  # avoid duplicate keyword arguments
                self._display_auth_status(profile, verbose=verbose, **auth)
        else:
            self._display_unauth_status(profile)

    def _display_auth_status(
        self,
        profile: Profile,
        *,
        verbose: bool = False,
        created_at: str,
        access_token: str,
        expires_in: int,
        token_type: str,
        scope: str,
        **kwargs,  # noqa: ARG002
    ):
        url = profile.extract_url()
        is_expired, time_diff = self._calculate_expiration(created_at, expires_in)
        status = '[red]✗[/red] Logged out of' if is_expired else '[green]✓[/green] Logged in to'
        status_label = f'expired {time_diff} ago' if is_expired else f'expires in {time_diff}'
        hidden_token = f'{access_token[:4]}{"*" * 24 }{access_token[-4:]}'

        console = Console()
        console.print(f'{profile.name}', style='magenta' if profile.is_active else '')
        console.print(f'  {status} [cyan]{url.tenant} ({str(url.env).upper()})[/cyan]')
        console.print(f'  - Token: [bold]{hidden_token}[/bold] {"(expired)" if is_expired else ""}')
        if verbose:
            console.print(f'  - Status: {status_label}')
            console.print(f'  - Scope: [bold]{scope}[/bold]')
            console.print(f'  - Token type: [bold]{token_type}[/bold]')
            console.print(f'  - Last updated: {datetime.fromisoformat(created_at).strftime(DATE_FORMAT)}')

    def _display_unauth_status(self, profile: Profile):
        url = profile.extract_url()
        console = Console()
        console.print(f'{profile.name}', style='magenta' if profile.is_active else '')
        console.print(f'  [red]✗[/red] Not logged in to [cyan]{url.tenant} ({str(url.env).upper()})[/cyan]')

    def _calculate_expiration(self, issued_at: str, expires_in: int) -> tuple[bool, str]:
        expiration_time = datetime.fromisoformat(issued_at) + timedelta(seconds=expires_in)
        current_time = datetime.now()

        if current_time >= expiration_time:
            return True, _friendly_time((current_time - expiration_time).total_seconds())  # expired at ...
        else:
            return False, _friendly_time((expiration_time - current_time).total_seconds())  # expires in ...


class AuthLoginCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='login',
            help='Use OAuth2 setup to retrieve or refresh access tokens',
            callback=self.login,
            params=[
                click.Option(
                    ['-r', '--refresh'],
                    is_flag=True,
                    default=False,
                    type=bool,
                    help='Force a refresh of the access token',
                ),
            ],
        )

    def login(self, refresh: bool):
        profile = get_active_profile()
        if not profile:
            return

        console = Console()
        if refresh:
            path = pathlib.Path(profile.oauth_path) if profile.oauth_path else None
            if path and path.exists():
                with path.open('r') as file:
                    oauth = json.load(file)
                    client_id = oauth.get('client_id') or oauth.get('clientId', '')
                    client_secret = oauth.get('client_secret') or oauth.get('clientSecret', '')
                self._login_with_oauth(profile, client_id, client_secret)
            elif profile.client_id and profile.client_secret:
                self._login_with_oauth(profile, profile.client_id, profile.client_secret)
            else:
                console.print('[red]✗[/red] No client ID or client secret found')
        else:
            path = HOME_DIR / f'{profile.name}_auth.json'
            if path.exists():
                with path.open('r') as file:
                    auth = json.load(file)

                    # NOTE: ideally, I should check the expiration time of the decoded token and
                    # make a decision based on that. However, for simplicity, I am using the
                    # the (local) time that the token was issued.
                    expiration_time = datetime.fromisoformat(auth['created_at']) + timedelta(seconds=auth['expires_in'])
                    current_time = datetime.now()

                    if current_time >= expiration_time:
                        console.print('[yellow]⚠[/yellow] Token has expired; refreshing token...')
                        self.login(refresh=True)
                    else:
                        expires_in = (expiration_time - current_time).total_seconds()
                        console.print(
                            f'[green]✓[/green] Token is still valid and will expire in about {_friendly_time(expires_in)}.'
                            '\nUse the [green]--refresh[/green] flag to force a token refresh'
                        )
            else:
                self.login(refresh=True)

    def _login_with_oauth(self, profile: Profile, client_id: str, client_secret: str):
        oauth = {'client_id': client_id, 'client_secret': client_secret}
        config = Config(
            base_url=profile.base_url,
            oauth=oauth,
            timeout=profile.timeout,
            max_retries=profile.max_retries,
            retry_interval=profile.retry_interval,
            logger=profile.logger,
        )

        console = Console()
        try:
            access_token = OAuth(oauth).retrieve_token(config)

            path = HOME_DIR / f'{profile.name}_auth.json'
            with path.open('w') as file:
                now = datetime.now().isoformat()
                json.dump({'profile': profile.name, 'created_at': now, **access_token.__dict__}, file, sort_keys=False)

            url = profile.extract_url()
            console.print(f'[green]✓[/green] Logged in to [cyan]{url.tenant} ({str(url.env).upper()})[/cyan]')
        except SparkApiError as err:
            if err.status == 401:
                console.print('[red]✗[/red] failed to log in to Spark due to invalid credentials')
            else:
                console.print(f'[red]✗[/red] failed to log in to Spark API...\nMore details: \n{err}')
        except Exception:
            console.print('[red]✗[/red] cannot to log in to Spark API')

    def _extract_creds(self):
        pass


class AuthLogoutCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='logout',
            help='Remove the authentication token for the active profile',
            options_metavar='',
            add_help_option=False,
            callback=self.logout,
        )

    def logout(self):
        profile = get_active_profile()
        url = profile.extract_url()
        tenant = f'[cyan]{url.tenant} ({str(url.env).upper()})[/cyan]'

        path = HOME_DIR / f'{profile.name}_auth.json'
        if path.exists():
            path.unlink()
            Console().print(f'[green]✓[/green] Logged out of {tenant}')
        else:
            Console().print(f'[red]✗[/red] No active session for {tenant}')


auth_cmd.add_command(AuthLoginCommand())
auth_cmd.add_command(AuthLogoutCommand())
auth_cmd.add_command(AuthStatusCommand())



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Generic, Optional, TypeVar, cast

from httpx import Headers, Request, Response

__all__ = ['SparkError', 'SparkSdkError', 'RetryTimeoutError', 'SparkApiError', 'ErrorMessage', 'ApiErrorCause']


class SparkError(Exception):
    """
    Base class for all SDK-related errors.

    Our exception hierarchy:
    * SparkError
        - SparkSdkError
            + RetryTimeoutError
        - SparkApiError
            + BadRequestError
            + UnauthorizedError
            + ForbiddenError
            + NotFoundError
            + ConflictError
            + UnsupportedMediaTypeError
            + UnprocessableEntityError
            + RateLimitError
            + InternalServerError
            + ServiceUnavailableError
            + GatewayTimeoutError
            + UnknownApiError
    """

    def __init__(self, message: str, cause: Optional[Any] = None):
        super().__init__()
        self.name = self.__class__.__name__
        self.message = message
        self.cause = cause

    def __str__(self):
        error, details = f'{self.name}: {self.message}', self.details
        return f'{error} ({details})' if details else error

    def __repr__(self):
        return f'<{self.name}>'

    def to_dict(self) -> Dict[str, Any]:
        cause = self.cause
        if isinstance(cause, ApiErrorCause):
            cause = {'request': cause.request.__dict__, 'response': cause.response.__dict__}
        elif isinstance(cause, Exception):
            cause = str(cause)
        return {'name': self.name, 'message': self.message, 'cause': cause}

    @property
    def details(self) -> str:
        if isinstance(self.cause, ApiErrorCause):
            req, resp = self.cause.request, self.cause.response
            return str({'request': req.__dict__, 'response': resp and resp.__dict__ or None})
        if isinstance(self.cause, (dict, Exception)):
            return str(self.cause)
        if isinstance(self.cause, str):
            return self.cause
        return ''

    @staticmethod
    def sdk(message: str, cause: Optional[Any] = None) -> SparkSdkError:
        return SparkSdkError(ErrorMessage(message, cause))

    @staticmethod
    def api(status: int, error: Dict[str, Any]) -> SparkApiError:
        return SparkApiError.when(status, ErrorMessage.from_dict(error))


class SparkSdkError(SparkError):
    """
    Base class for SDK-specific errors.

    Usually thrown when an argument fails to comply with the expected format.
    Because it's a client-side error, it will include in the majority of cases
    the invalid entry as `cause`.

    If the argument causes another process to fail, the `cause` will be referencing
    that thrown `Exception`. The `timestamp` helps to define a hierarchy of errors since
    the `cause` might be a `SparkError` as well.
    """

    def __init__(self, error: ErrorMessage):
        super().__init__(error.message, error.cause)
        self.timestamp = datetime.now().isoformat()

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), 'timestamp': self.timestamp}


class RetryTimeoutError(SparkSdkError):
    """Raised when the maximum number of retries is reached."""

    def __init__(self, message: str, *, cause: Optional[Any] = None, retries: int = 0, interval: float = 0.0):
        super().__init__(ErrorMessage(message, cause))
        self.retries = retries
        self.interval = interval

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), 'retries': self.retries, 'interval': self.interval}


class SparkApiError(SparkError):
    """
    Base class for errors related to the API.

    When attempting to communicate with the API, the SDK will wrap any sort of failure
    (any error during the round trip) into `SparkApiError`. The `status` is the HTTP
    status code of the response, and the `request_id`, a unique identifier of the request.
    """

    def __init__(self, error: ErrorMessage, status: Optional[int] = None):
        cause = ApiErrorCause.from_dict(error.cause) if isinstance(error.cause, dict) else None
        super().__init__(f"{status or ''} {error.message}".strip(), cause)
        self.status = status

    @property
    def request_id(self) -> str:
        if self.cause and isinstance(self.cause, ApiErrorCause):
            return self.cause.request.headers.get('x-request-id', '')
        return ''  # Note: this should never happen if thrown by the SDK

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), 'status': self.status}

    @staticmethod
    def when(status: int, error: ErrorMessage) -> SparkApiError:
        if status == 400:
            return BadRequestError(error, status)
        elif status == 401:
            return UnauthorizedError(error, status)
        elif status == 403:
            return ForbiddenError(error, status)
        elif status == 404:
            return NotFoundError(error, status)
        elif status == 409:
            return ConflictError(error, status)
        elif status == 415:
            return UnsupportedMediaTypeError(error, status)
        elif status == 422:
            return UnprocessableEntityError(error, status)
        elif status == 429:
            return RateLimitError(error, status)
        elif status == 500:
            return InternalServerError(error, status)
        elif status == 503:
            return ServiceUnavailableError(error, status)
        elif status == 504:
            return GatewayTimeoutError(error, status)
        else:
            return UnknownApiError(error)

    @staticmethod
    def to_cause(request: Request, response: Response) -> dict[str, Any]:
        return {
            'request': {
                'url': str(request.url),
                'method': request.method,
                'headers': request.headers,
                'body': request.content,
            },
            'response': {
                'headers': response.headers,
                'body': response.text,
                'raw': response.content,
            },
        }

    @staticmethod
    def no_response(request: Request) -> dict[str, Any]:
        return {
            'request': {
                'url': str(request.url),
                'method': request.method,
                'headers': request.headers,
                'body': request.content,
            }
        }


class BadRequestError(SparkApiError):
    status = 400


class UnauthorizedError(SparkApiError):
    status = 401


class ForbiddenError(SparkApiError):
    status = 403


class NotFoundError(SparkApiError):
    status = 404


class ConflictError(SparkApiError):
    status = 409


class UnsupportedMediaTypeError(SparkApiError):
    status = 415


class UnprocessableEntityError(SparkApiError):
    status = 422


class RateLimitError(SparkApiError):
    status = 429


class InternalServerError(SparkApiError):
    status = 500


class ServiceUnavailableError(SparkApiError):
    status = 503


class GatewayTimeoutError(SparkApiError):
    status = 504


class UnknownApiError(SparkApiError):
    status = None


TReq = TypeVar('TReq')
TResp = TypeVar('TResp')


class TRequest(Generic[TReq]):
    def __init__(self, url: str, method: str, headers: Headers, body: Optional[TReq]):
        self.url = url
        self.method = method
        self.headers = headers
        self.body = body


class TResponse(Generic[TResp]):
    def __init__(self, headers: Headers, body: TResp, raw: str):
        self.headers = headers
        self.body = body
        self.raw = raw


class ErrorMessage:
    def __init__(self, message: str, cause: Optional[Any] = None):
        self.message = message
        self.cause = cause

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> ErrorMessage:
        message = cast(str, data.get('message'))
        cause = cast(Optional[Any], data.get('cause'))
        return ErrorMessage(message, cause)


class ApiErrorCause(Generic[TReq, TResp]):
    def __init__(self, request: TRequest[TReq], response: Optional[TResponse[TResp]] = None):
        self.request = request
        self.response = response

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> ApiErrorCause:
        req, res = data.get('request', {}), data.get('response')
        request = TRequest(req.get('url', ''), req.get('method', ''), req.get('headers', {}), req.get('body', ''))
        if not res:
            return ApiErrorCause(request)
        return ApiErrorCause(request, TResponse(res['headers'], res['body'], res['raw']))



import gzip
import json
import time
import zlib
from datetime import datetime
from typing import Any, BinaryIO, Dict, List, Mapping, Optional, Tuple, Union

from .._constants import SPARK_SDK
from .._errors import RetryTimeoutError, SparkError
from .._utils import DateUtils, StringUtils, get_retry_timeout
from ._base import ApiResource, HttpResponse, Uri, UriParams
from ._transforms import TransformParams

__all__ = ['Services', 'ServiceExecuted']


class Services(ApiResource):
    @property
    def compilation(self):
        return Compilation(self.config)

    def create(
        self,
        name: str,
        *,
        folder: str,
        file: BinaryIO,
        file_name: Optional[str] = None,
        draft_name: Optional[str] = None,
        versioning: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        track_user: Optional[bool] = False,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        compiled = self.compile(
            folder=folder,
            service=name,
            file=file,
            file_name=file_name,
            versioning=versioning,
            start_date=start_date,
            end_date=end_date,
            max_retries=max_retries,
            retry_interval=retry_interval,
        )
        upload = compiled['upload'].get('response_data', {})

        published = self.publish(
            folder=folder,
            service=name,
            file_id=upload.get('original_file_documentid'),
            engine_id=upload.get('engine_file_documentid'),
            draft_name=draft_name,
            versioning=versioning,
            start_date=start_date,
            end_date=end_date,
            track_user=track_user,
        )
        return {**compiled, 'publication': published.data}

    def compile(
        self,
        *,
        folder: str,
        service: str,
        file: BinaryIO,
        file_name: Optional[str] = None,
        versioning: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        with self.compilation as compilation:
            upload = compilation.initiate(
                folder=folder,
                service=service,
                file=file,
                file_name=file_name,
                versioning=versioning,
                start_date=start_date,
                end_date=end_date,
            )

            status = compilation.get_status(
                job_id=isinstance(upload.data, dict)
                and upload.data.get('response_data', {}).get('nodegen_compilation_jobid')
                or '',  # NOTE: this should never happen (only bypassing type checker)
                folder=folder,
                service=service,
                max_retries=max_retries,
                retry_interval=retry_interval,
            )
            return {'upload': upload.data, 'compilation': status.data}

    def publish(
        self,
        folder: str,
        service: str,
        file_id: str,
        engine_id: str,
        draft_name: Optional[str] = None,
        versioning: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        track_user: Optional[bool] = False,
    ):
        startdate, enddate = DateUtils.parse(start_date, end_date)
        uri = Uri.validate(UriParams(folder, service))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint='publish')
        params = {
            'draft_service_name': draft_name or uri.service,
            'effective_start_date': startdate.isoformat(),
            'effective_end_date': enddate.isoformat(),
            'original_file_documentid': file_id,
            'engine_file_documentid': engine_id,
            'version_difference': versioning or 'minor',
            'should_track_user_action': track_user,
        }

        response = self.request(url, method='POST', body={'request_data': params})
        version_id = isinstance(response.data, dict) and response.data.get('response_data', {}).get('version_id')
        if version_id:
            self.logger.info(f'service published with version id <{version_id}>')
        return response

    def execute(
        self,
        uri: Union[str, UriParams],
        *,
        response_format: Optional[str] = None,
        encoding: Optional[str] = None,  # 'gzip' | 'deflate'
        # data for calculations
        inputs: Union[None, str, Dict[str, Any], List[Any]] = None,  # TODO: support `pandas.DataFrame`
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
        # extra metadata if needed
        extras: Optional[Mapping[str, Any]] = None,
    ):
        uri = Uri.validate(uri)

        executable = _ExecuteInputs(inputs)
        metadata = _ExecuteMeta(
            uri,
            is_batch=executable.is_batch,
            active_since=active_since,
            source_system=source_system,
            correlation_id=correlation_id,
            call_purpose=call_purpose,
            compiler_type=compiler_type,
            subservices=subservices,
            debug_solve=debug_solve,
            downloadable=downloadable,
            echo_inputs=echo_inputs,
            tables_as_array=tables_as_array,
            selected_outputs=selected_outputs,
            outputs_filter=outputs_filter,
            extras=extras,
        )

        if executable.is_batch:
            url = Uri.of(uri.pick('public'), base_url=self.config.base_url.full, version='api/v4', endpoint='execute')
            body = {'inputs': executable.inputs, **metadata.values}
        else:
            endpoint = '' if uri.version_id or uri.service_id else 'execute'
            url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=endpoint)
            body = {'request_data': {'inputs': executable.inputs}, 'request_meta': metadata.values}

        if encoding:
            content, headers = self.__encode(data=body, encoding=encoding)
            response = self.request(url, method='POST', content=content, headers=headers)
        else:
            response = self.request(url, method='POST', body=body)
        return ServiceExecuted(response, executable.is_batch, response_format or 'alike')

    def transform(
        self,
        uri: Union[str, UriParams],
        *,
        # data for calculations
        inputs: Any,  # required
        using: Union[str, Dict[str, str], None] = None,
        api_version: str = 'v3',  # 'v3' | 'v4'
        encoding: Optional[str] = None,  # 'gzip' | 'deflate'
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Only for api/v3
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
        # extra metadata if needed
        extras: Optional[Mapping[str, Any]] = None,
    ):
        uri = Uri.validate(uri)

        metadata = _ExecuteMeta(
            uri,
            is_batch=api_version == 'v4',
            active_since=active_since,
            source_system=source_system,
            correlation_id=correlation_id,
            call_purpose=call_purpose,
            compiler_type=compiler_type,
            subservices=subservices,
            debug_solve=debug_solve,
            downloadable=downloadable,
            echo_inputs=echo_inputs,
            tables_as_array=tables_as_array,
            selected_outputs=selected_outputs,
            outputs_filter=outputs_filter,
            extras=extras,
        )

        if isinstance(using, str):
            # (legacy) using transform name for Documents saved under Options > /apps/transforms
            endpoint = f'transforms/{using}/for/folders/{uri.folder}/services/{uri.service}'
        else:
            using = using or {}  # use 'folder' and 'service' as fallback if using is wrong
            transform = TransformParams(using.get('folder', uri.folder), using.get('service', uri.service))
            endpoint = f'transforms/{transform.folder}/{transform.name}/for/{uri.folder}/{uri.service}'

        url = Uri.of(base_url=self.config.base_url.full, version='api/v4', endpoint=endpoint)

        if encoding:
            content, headers = self.__encode(data=inputs or {}, encoding=encoding, extras=metadata.as_header)
            response = self.request(url, method='POST', content=content, headers=headers)
        else:
            response = self.request(url, method='POST', body=inputs or {}, headers=metadata.as_header)
        return response

    def validate(
        self,
        uri: Union[str, UriParams],
        *,
        # data for validations
        inputs: Union[None, str, Dict[str, Any]] = None,
        validation_type: Optional[str] = None,  # 'dynamic' | 'static'
        # Metadata for validations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
        # extra metadata if needed
        extras: Optional[Mapping[str, Any]] = None,
    ):
        uri = Uri.validate(uri)
        validation_type = StringUtils.is_not_empty(validation_type) and str(validation_type).lower() or None

        executable = _ExecuteInputs(inputs)
        metadata = _ExecuteMeta(
            uri,
            is_batch=False,
            active_since=active_since,
            source_system=source_system,
            correlation_id=correlation_id,
            call_purpose=call_purpose,
            compiler_type=compiler_type,
            subservices=subservices,
            debug_solve=debug_solve,
            downloadable=downloadable,
            echo_inputs=echo_inputs,
            tables_as_array=tables_as_array,
            selected_outputs=selected_outputs,
            outputs_filter=outputs_filter,
            extras=extras,
        )
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint='validation')
        body = {
            'request_data': {'inputs': executable.inputs},
            'request_meta': {
                **metadata.values,
                'validation_type': 'dynamic' if validation_type == 'dynamic' else 'default_values',
            },
        }

        return self.request(url, method='POST', body=body)

    def get_schema(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        version_id: Optional[str] = None,
    ):
        uri = Uri.validate(Uri.to_params(uri) if uri else UriParams(folder, service, version_id=version_id))
        base, folder, service = self.config.base_url, uri.folder, uri.service
        if StringUtils.is_not_empty(uri.version_id):
            url = Uri.partial(f'GetEngineDetailByVersionId/versionid/{uri.version_id}', base_url=base.full)
            return self.request(url, method='POST')
        else:
            url = Uri.of(base_url=base.value, version='api/v1', endpoint=f'product/{folder}/engines/get/{service}')
            return self.request(url, method='GET')

    def get_metadata(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        service_id: Optional[str] = None,
        version_id: Optional[str] = None,
        proxy: Optional[str] = None,
        public: Optional[bool] = False,
    ):
        uri = Uri.validate(uri or UriParams(folder, service, service_id, None, version_id, proxy, public))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint='metadata')

        response = self.request(url)
        return ServiceExecuted(response, False, 'original')

    def get_versions(
        self, uri: Union[None, str, UriParams] = None, *, folder: Optional[str] = None, service: Optional[str] = None
    ):
        uri = Uri.validate(uri or UriParams(folder, service))
        endpoint = f'product/{uri.folder}/engines/getversions/{uri.service}'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)

        response = self.request(url)
        return response.copy_with(data=response.data.get('data', []) if isinstance(response.data, dict) else [])

    def get_swagger(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        version_id: Optional[str] = None,
        downloadable: Optional[bool] = False,
        subservice: str = 'All',
    ):
        uri = Uri.validate(uri or UriParams(folder, service))
        endpoint = f'downloadswagger/{subservice}/{downloadable}/{version_id or ""}'
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=endpoint)

        return self.request(url, method='GET')

    def search(
        self,
        *,
        page: int = 1,
        limit: int = -1,
        sort: str = 'name1_co',
        query: Optional[List[Any]] = None,
        fields: Optional[List[str]] = None,
        **params: Any,
    ):
        uri = Uri.of(base_url=self.config.base_url.full, endpoint='services/search')
        search_params = {
            'page': page,
            'page_size': limit,
            'sort': sort,
            'search': query or [],
            'fields': fields or ['id', 'foldername', 'filename', 'version', 'modifiedDate'],
            **params,  # other query parameters
        }

        return self.request(uri, method='POST', body={'request_data': search_params})

    def download(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        version_id: Optional[str] = None,
        file_name: Optional[str] = None,
        type: Optional[str] = None,  # 'original' or 'configured'
    ):
        uri = Uri.validate(uri or UriParams(folder, service))
        endpoint = f'product/{uri.folder}/engines/{uri.service}/download/{version_id or ""}'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)
        params = {'filename': file_name or '', 'type': 'withmetadata' if type == 'configured' else ''}

        return self.request(url, params=params)

    def recompile(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        version_id: Optional[str] = None,
        compiler: Optional[str] = None,
        release_notes: Optional[str] = None,
        label: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        upgrade: Optional[str] = None,  # 'major' | 'minor' | 'patch'
        tags: Union[None, str, List[str]] = None,
    ):
        uri = Uri.validate(uri or UriParams(folder, service, version_id=version_id))
        url = Uri.of(uri.pick('folder', 'service'), base_url=self.config.base_url.full, endpoint='recompileNodgen')
        startdate, enddate = DateUtils.parse(start_date, end_date)
        data = {
            'versionId': uri.version_id,
            'upgradeType': upgrade or 'patch',
            'neuronCompilerVersion': compiler or 'StableLatest',
            'releaseNotes': release_notes or f'Recompiled via {SPARK_SDK}',
            'label': label,
            'effectiveStartDate': startdate.isoformat(),
            'effectiveEndDate': enddate.isoformat(),
            'tags': StringUtils.join(tags),
        }

        return self.request(url, method='POST', body={'request_data': data})

    def delete(
        self, uri: Union[None, str, UriParams] = None, *, folder: Optional[str] = None, service: Optional[str] = None
    ):
        uri = Uri.validate(uri or UriParams(folder, service))
        endpoint = f'product/{uri.folder}/engines/delete/{uri.service}'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)

        return self.request(url, method='DELETE')

    def __encode(
        self,
        *,
        data: Any,
        encoding: str = 'gzip',
        content_type: str = 'application/json',
        extras: Mapping[str, str] = {},
    ) -> Tuple[bytes, Dict[str, str]]:
        headers = {'Content-Type': content_type, 'Content-Encoding': encoding, 'Accept-Encoding': encoding, **extras}
        if encoding == 'gzip':
            return gzip.compress(json.dumps(data).encode('utf-8')), headers
        if encoding == 'deflate':
            return zlib.compress(json.dumps(data).encode('utf-8')), headers
        else:
            raise SparkError.sdk(f'encoding "{encoding}" is not supported', {'encoding': encoding})


class ServiceExecuted(HttpResponse):
    def __init__(self, response: HttpResponse, is_batch: bool, format: str = 'alike'):
        if format == 'original' or is_batch:
            data = response.data
        else:
            resp_data = response.data.get('response_data', {}) if isinstance(response.data, dict) else {}
            resp_meta = response.data.get('response_meta', {}) if isinstance(response.data, dict) else {}
            data = {
                'outputs': [resp_data.get('outputs')],
                'process_time': [resp_meta.get('process_time')],
                'warnings': [resp_data.get('warnings')],
                'errors': [resp_data.get('errors')],
                'service_chain': [resp_meta.get('service_chain')],
                'service_id': resp_meta.get('service_id'),
                'version_id': resp_meta.get('version_id'),
                'version': resp_meta.get('version'),
                'call_id': resp_meta.get('call_id'),
                'compiler_version': resp_meta.get('compiler_version'),
                'correlation_id': resp_meta.get('correlation_id'),
                'request_timestamp': resp_meta.get('request_timestamp'),
            }
        super().__init__(
            response.status, data, response.buffer, response.headers, response.raw_request, response.raw_response
        )


class _ExecuteInputs:
    def __init__(self, data: Union[None, str, Dict[str, Any], List[Any]] = None):
        if data is None or (isinstance(data, list) and len(data) == 0):
            data = {}
        if StringUtils.is_not_empty(data):
            data = json.loads(str(data))

        self.inputs = data
        if isinstance(data, dict):
            self.length = 1
            self.is_batch = False
        elif isinstance(data, list):
            self.length = len(data)
            self.is_batch = True
        else:
            message = 'invalid data format\nexpected input data formats are string, dict or a list'
            raise SparkError.sdk(message, data)


class _ExecuteMeta:
    __COMPILER_TYPES = ('neuron', 'type3', 'type2', 'type1', 'xconnector')

    def __init__(
        self,
        uri: UriParams,
        is_batch: bool,
        *,
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3 (legacy)
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
        # extra metadata if needed
        extras: Optional[Mapping[str, Any]] = None,
    ):
        self._uri = uri
        self._is_batch = is_batch
        self._active_since = active_since
        self._source_system = source_system or SPARK_SDK
        self._correlation_id = correlation_id

        self._call_purpose = (
            call_purpose
            if StringUtils.is_not_empty(call_purpose)
            else 'Sync Batch Execution'
            if is_batch
            else 'Single Execution'
        )
        self._compiler_type = (
            str(compiler_type).capitalize()
            if StringUtils.is_not_empty(compiler_type) and str(compiler_type).lower() in self.__COMPILER_TYPES
            else 'Neuron'
        )

        self._subservices = StringUtils.join(subservices)
        self._debug_solve = debug_solve
        self._downloadable = downloadable
        self._echo_inputs = echo_inputs
        self._tables_as_array = StringUtils.join(tables_as_array)
        self._selected_outputs = StringUtils.join(selected_outputs)
        self._outputs_filter = outputs_filter
        self._extras = extras or {}

    @property
    def values(self) -> Dict[str, Any]:
        if self._is_batch:
            service_uri = self._uri.pick('folder', 'service', 'version').encode(long=False)
            values = {
                'service': self._uri.service_id or service_uri or None,
                'version_id': self._uri.version_id,
                'version_by_timestamp': self._active_since,
                'subservice': self._subservices,
                'output': self._selected_outputs,
                'call_purpose': self._call_purpose,
                'source_system': self._source_system,
                'correlation_id': self._correlation_id,
                # extra metadata
                **self._extras,
            }
        else:
            values = {
                # URI locator via metadata (v3 also supports URI in url path)
                'service_id': self._uri.service_id,
                'version_id': self._uri.version_id,
                'version': self._uri.version,
                # v3 expects extra metadata
                'transaction_date': self._active_since,
                'call_purpose': self._call_purpose,
                'source_system': self._source_system,
                'correlation_id': self._correlation_id,
                'array_outputs': self._tables_as_array,
                'compiler_type': self._compiler_type,
                'debug_solve': self._debug_solve,
                'excel_file': self._downloadable,
                'requested_output': self._selected_outputs,
                'requested_output_regex': self._outputs_filter,
                'response_data_inputs': self._echo_inputs,
                'service_category': self._subservices,
                # extra metadata
                **self._extras,
            }

        return {k: v for k, v in values.items() if v is not None}

    @property
    def as_header(self) -> Dict[str, str]:
        # NOTE: this has to be a single line string: "'{\"call_purpose\":\"Single Execution\"}'"
        value = json.dumps(self.values, separators=(',', ':'))
        return {'x-meta' if self._is_batch else 'x-request-meta': "'{}'".format(value)}


class Compilation(ApiResource):
    def initiate(
        self,
        folder: str,
        service: str,
        file: BinaryIO,
        file_name: Optional[str] = None,
        versioning: Optional[str] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
    ):
        startdate, enddate = DateUtils.parse(start_date, end_date)
        uri = Uri.validate(UriParams(folder, service))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint='upload')

        metadata = {
            'request_data': {
                'effective_start_date': startdate.isoformat(),
                'effective_end_date': enddate.isoformat(),
                'version_difference': versioning or 'minor',
            }
        }
        form = {'engineUploadRequestEntity': json.dumps(metadata)}
        files = {'serviceFile': (file_name or f'{uri.service}.xlsx', file)}

        response = self.request(url, method='POST', form=form, files=files)
        if isinstance(response.data, dict) and response.data.get('response_data'):
            doc_id = response.data.get('response_data', {}).get('original_file_documentid')
            self.logger.info(f'service file uploaded <{doc_id}>')
        return response

    def get_status(
        self,
        folder: str,
        service: str,
        job_id: str,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        throwable: bool = True,
    ):
        """Polls the compilation job status until it's completed or timed out."""
        max_retries = max_retries or self.config.max_retries
        retry_interval = retry_interval or self.config.retry_interval

        uri = Uri.validate(UriParams(folder, service))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=f'getcompilationprogess/{job_id}')

        retries = 0
        response = self.request(url)
        while True:
            response_data = isinstance(response.data, dict) and response.data.get('response_data', {}) or {}
            progress, status = response_data.get('progress', 0), response_data.get('status')

            if progress == 100 and status == 'Success':
                return response

            if progress < 100 and retries < max_retries:
                self.logger.info(f'waiting for compilation job to complete - {progress}%')

                retries += 1
                time.sleep(get_retry_timeout(retries, retry_interval))
                response = self.request(url)
            else:
                err_msg = f'compilation job status check timed out after {retries} attempts'
                if throwable:
                    self.logger.error(err_msg)
                    raise RetryTimeoutError(err_msg, retries=retries, interval=retry_interval)
                self.logger.warning(err_msg)
                return response



__all__ = [
    'ENV_VARS',
    'DEFAULT_MAX_RETRIES',
    'DEFAULT_RETRY_INTERVAL',
    'DEFAULT_TIMEOUT_IN_MS',
    'RETRY_RANDOMIZATION_FACTOR',
    'DEFAULT_LOGGER_FORMAT',
    'DEFAULT_LOGGER_DATEFMT',
    'SPARK_SDK',
]

DEFAULT_MAX_RETRIES = 2
DEFAULT_RETRY_INTERVAL = 1  # 1 second
DEFAULT_TIMEOUT_IN_MS = 60000  # 60 seconds
RETRY_RANDOMIZATION_FACTOR = 1.5
DEFAULT_LOGGER_FORMAT = '[%(name)s] %(asctime)s %(levelname)s - %(message)s'
DEFAULT_LOGGER_DATEFMT = '%m/%d/%Y, %I:%M:%S %p'
SPARK_SDK = 'Spark Python SDK'


class ENV_VARS:
    BASE_URL = 'CSPARK_BASE_URL'
    API_KEY = 'CSPARK_API_KEY'
    BEARER_TOKEN = 'CSPARK_BEARER_TOKEN'
    CLIENT_ID = 'CSPARK_CLIENT_ID'
    CLIENT_SECRET = 'CSPARK_CLIENT_SECRET'
    OAUTH_PATH = 'CSPARK_OAUTH_PATH'



from typing import Any, BinaryIO, Dict, List, Mapping, Optional, Union

from cspark.sdk import Services as SdkServices
from cspark.sdk import Uri, UriParams

from ._base import HybridResource

__all__ = ['Services']


class Services(HybridResource):
    def upload(self, file: BinaryIO, file_name: Optional[str] = None):
        url = Uri.of(base_url=self.config.base_url.value, endpoint='upload')
        return self.request(url, method='POST', files={'file': (file_name or 'package.zip', file)})

    def execute(
        self,
        uri: Union[str, UriParams],
        *,
        inputs: Union[None, str, Dict[str, Any], List[Any]] = None,
        encoding: Optional[str] = None,
        response_format: Optional[str] = None,
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3
        debug_solve: Optional[bool] = None,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
    ):
        with SdkServices(self.config) as services:
            return services.execute(
                uri,
                inputs=inputs,
                response_format=response_format,
                encoding=encoding,
                subservices=subservices,
                active_since=active_since,
                source_system=source_system,
                correlation_id=correlation_id,
                call_purpose=call_purpose,
                compiler_type=compiler_type,
                debug_solve=debug_solve,
                echo_inputs=echo_inputs,
                tables_as_array=tables_as_array,
                selected_outputs=selected_outputs,
                outputs_filter=outputs_filter,
            )

    def validate(
        self,
        uri: Union[str, UriParams],  # only version_id or service_id formats are supported
        *,
        service_id: Optional[str] = None,
        version_id: Optional[str] = None,
        inputs: Union[None, Dict[str, Any]] = None,  # data for validations
        validation_type: Optional[str] = None,  # 'dynamic' | 'static'
        metadata: Optional[Mapping[str, Any]] = None,  # extra metadata for validations
    ):
        uri = Uri.validate(uri or UriParams(service_id=service_id, version_id=version_id))
        url = Uri.of(base_url=self.config.base_url.full, endpoint='validation')
        body = {
            'request_data': {'inputs': inputs},
            'request_meta': {
                **(metadata or {}),
                'service_id': uri.service_id,
                'version_id': uri.version_id,
                'version': uri.version,
                'validation_type': 'dynamic' if validation_type == 'dynamic' else 'default_values',
            },
        }

        return self.request(url, method='POST', body=body)

    def get_metadata(
        self,
        uri: Union[str, UriParams],  # only version_id or service_id formats are supported
        *,
        inputs: Union[None, Dict[str, Any]] = None,
        service_id: Optional[str] = None,
        version_id: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,  # extra metadata fields
    ):
        uri = Uri.validate(uri or UriParams(service_id=service_id, version_id=version_id))
        url = Uri.of(base_url=self.config.base_url.full, endpoint='metadata')
        body = {
            'request_data': {'inputs': inputs},
            'request_meta': {
                **(metadata or {}),
                'service_id': uri.service_id,
                'version_id': uri.version_id,
            },
        }

        return self.request(url, method='POST', body=body)



from ._base import *
from ._batches import *
from ._folders import *
from ._health import *
from ._history import *
from ._impex import *
from ._oauth2 import *
from ._services import *
from ._transforms import *



from dataclasses import dataclass
from json import dumps, loads
from typing import Any, List, Mapping, Optional, Union

from .._config import Config
from .._errors import SparkError
from .._validators import Validators
from ._base import ApiResource, Uri

__all__ = ['Transforms', 'Transform']


@dataclass(frozen=True)
class Transform:
    schema: Optional[str] = None
    api_version: Optional[str] = None
    inputs: Optional[str] = None
    outputs: Optional[str] = None
    extras: Optional[Mapping[str, Any]] = None


@dataclass(frozen=True)
class TransformParams:
    folder: Optional[str] = None
    name: Optional[str] = None

    def validate(self) -> None:
        if not self.folder or not self.name:
            raise SparkError.sdk('transform folder and name are required', self.__dict__)

    def to_uri(self) -> str:
        self.validate()
        return f'{self.folder}/{self.name}'


class Transforms(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.full, 'version': 'api/v4'}

    def list(
        self,
        folder: str,
        *,
        page: int = 1,
        size: int = 100,
        sort: str = '-updatedAt',
        search: Optional[List[Any]] = None,
    ):
        url = Uri.of(None, endpoint=f'transform/list/{folder}', **self._base_uri)
        request_data = {'page': page, 'pageSize': size, 'sort': sort, 'search': search or []}

        return self.request(url, method='POST', body={'request_data': request_data})

    def validate(self, transform: Union[str, Transform]):
        url = Uri.of(None, endpoint=f'transform/validation', **self._base_uri)
        body = {'transform_content': self.__build(transform)}

        return self.request(url, method='POST', body=body)

    def get(self, uri: Optional[str] = None, *, folder: Optional[str] = None, name: Optional[str] = None):
        endpoint = f'transform/{uri or TransformParams(folder, name).to_uri()}'
        url = Uri.of(None, endpoint=endpoint, **self._base_uri)

        return self.request(url, method='GET')

    def save(self, *, folder: str, name: str, transform: Union[str, Transform]):
        url = Uri.of(None, endpoint=f'transform/{folder}/{name}', **self._base_uri)
        body = {'transform_content': self.__build(transform)}

        return self.request(url, method='POST', body=body)

    def delete(self, uri: Optional[str] = None, *, folder: Optional[str] = None, name: Optional[str] = None):
        endpoint = f'transform/{uri or TransformParams(folder, name).to_uri()}'
        url = Uri.of(None, endpoint=endpoint, **self._base_uri)

        return self.request(url, method='DELETE')

    def __build(self, value: Union[str, Transform]) -> str:
        value = Transform(**loads(value)) if isinstance(value, str) else value
        extras = value.extras or {}

        if value.schema and 'nodejs22' in value.schema.lower():
            return dumps({'transform_type': value.schema, 'transform_code': value.inputs, **extras})

        transform = dumps(
            {
                'transform_type': value.schema or 'JSONtransforms_v1.0.1',
                'target_api_version': value.api_version or 'v3',
                'input_body_transform': value.inputs,
                'output_body_transform': value.outputs,
                **extras,
            }
        )
        Validators.transform().validate(transform)

        return transform



from ._base import ApiResource

__all__ = ['Health', 'Platform']


class Health(ApiResource):
    def check(self):
        """Checks the health status or connectivity of a Spark environment."""
        return self.request(f'{self.config.base_url.value}/health', method='GET')

    def ok(self) -> bool:
        """Returns `True` if the Spark environment is healthy."""
        response = self.check()
        status = response.data.get('status', '').lower() if isinstance(response.data, dict) else ''
        return response.status == 200 and status == 'up'


class Platform(ApiResource):
    def get_config(self):
        """Returns Spark configuration for the current user."""
        url = f'{self.config.base_url.value}/api/v1/config/GetSparkConfiguration'
        response = self.request(url, method='GET')
        if isinstance(response.data, dict) and response.data.get('status') == 'Success':
            return response.copy_with(data=response.data.get('data'))
        return response



import json
import pathlib
import re
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Union

import click
import yaml
from cspark.sdk import BaseUrl, LoggerOptions, SparkError
from rich.console import Console

DATE_FORMAT = '%Y-%m-%d %I:%M:%S %p'
HOME_DIR = pathlib.Path.home() / '.cspark'
_PROFILE_PATH = HOME_DIR / 'profiles.yml'


class NoProfileError(click.UsageError):
    def __init__(self, ctx: Optional[click.Context] = None):
        msg = 'no profiles defined yet...\n'
        msg += 'Use `cspark init` to create and set an active profile.'
        super().__init__(msg, ctx=ctx)


@dataclass
class Profile:
    name: str
    is_active: bool
    created_at: str
    updated_at: str
    base_url: str
    api_key: Optional[str] = None
    token: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    oauth_path: Optional[str] = None
    timeout: Optional[float] = None
    max_retries: Optional[int] = None
    retry_interval: Optional[float] = None
    logger: Union[bool, LoggerOptions] = False

    def mask_auth(self, show: bool = False) -> dict[str, str]:
        def mask(s, count=24):
            return (s or '')[:4] + '*' * count + (s or '')[-4:]

        auth = dict[str, str]()
        if self.api_key:
            auth['api_key'] = self.api_key if show else mask(self.api_key, 12)
        if self.token:
            auth['token'] = self.token if show else mask(self.token)
        if self.oauth_path:
            auth['oauth'] = self.oauth_path
        elif self.client_id and self.client_secret:
            auth['oauth'] = (
                f'{self.client_id} / {self.client_secret}'
                if show
                else f'{mask(self.client_id)} / {mask(self.client_secret)}'
            )
        return auth

    def extract_url(self) -> BaseUrl:
        try:
            return BaseUrl.of(url=self.base_url)
        except SparkError as err:
            raise click.Abort(f'invalid base URL: {self.base_url}') from err

    def to_dict(self) -> dict:
        d = {
            'name': self.name,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'base_url': self.base_url,
            'auth': {},
            'timeout': self.timeout,
            'max_retries': self.max_retries,
            'retry_interval': self.retry_interval,
            'logger': self.logger.__dict__ if isinstance(self.logger, LoggerOptions) else False,
        }

        if self.api_key:
            d['auth']['api_key'] = self.api_key
        if self.token:
            d['auth']['token'] = self.token
        if self.client_id and self.client_secret:
            d['auth']['oauth'] = {'client_id': self.client_id, 'client_secret': self.client_secret}
        elif self.oauth_path:
            d['auth']['oauth'] = {'path': self.oauth_path}

        return {k: v for k, v in d.items() if v is not None}

    @staticmethod
    def from_dict(data: dict) -> 'Profile':
        now = datetime.now().isoformat()
        return Profile(
            name=data.get('name', ''),
            is_active=True,
            created_at=now,
            updated_at=now,
            base_url=data.get('base_url', ''),
            api_key=data.get('api_key'),
            token=data.get('token'),
            client_id=data.get('client_id'),
            client_secret=data.get('client_secret'),
            oauth_path=data.get('oauth_path'),
            timeout=data.get('timeout'),
            max_retries=data.get('max_retries'),
            retry_interval=data.get('retry_interval'),
            logger=LoggerOptions(**data.get('logger', {})) if data.get('logger') else False,
        )

    def to_config(self) -> dict:
        d = self.to_dict()
        del d['name']
        del d['created_at']
        del d['updated_at']

        path = HOME_DIR / f'{self.name}_auth.json'
        if path.exists():
            with path.open('r') as file:
                auth = json.load(file)
                d.update({'token': auth.get('access_token')})
                d.pop('auth', None)
        else:
            auth = d.pop('auth', {})
            oauth = auth.pop('oauth', {})
            d.update({'oauth': oauth['path']} if 'path' in oauth else oauth)  # reshape oauth
            d.update(auth)

        return d

    def _validate(self):
        if not self.name:
            raise click.UsageError('profile name is required')
        if not self.base_url:
            raise click.UsageError('base_url is required')
        if not self.api_key and not self.token and not self.client_id and not self.oauth_path:
            raise click.UsageError('auth is required')
        if self.client_id and not self.client_secret:
            raise click.UsageError('client_secret is required')


def create_profile():
    if not HOME_DIR.exists():
        HOME_DIR.mkdir()

    path = _PROFILE_PATH
    if path.exists():
        return

    with path.open('w') as file:
        yaml.dump({'version': '1.0', 'profile': '', 'accounts': []}, file, sort_keys=False)


def load_profiles(is_init: bool = False) -> List[Profile]:
    path = _PROFILE_PATH
    if not path.exists():
        create_profile()

    with path.open('r') as file:
        profile = yaml.load(file, Loader=yaml.FullLoader)

    version = profile.get('version', '1.0')
    active = profile.get('profile')
    accounts = profile.get('accounts', [])

    if re.match(r'^1\.\d+$', version):  # should match: 1.*
        profiles = [_parse_v1(account, active) for account in accounts]
        if not is_init and len(profiles) == 0:
            Console().print(
                '[red]ERROR: no profile has been set yet...[/red]'
                '\n💡 Use [green]cspark init[/green] to create and set an active profile.'
            )
            raise click.exceptions.Exit(1)
        return profiles
    raise ValueError(f'unsupported profile version: {version}')


def add_or_update_profile(updated: Profile):
    updated._validate()

    path = _PROFILE_PATH
    if not path.exists():
        create_profile()

    with path.open('r') as file:
        data = yaml.load(file, Loader=yaml.FullLoader)

    active_profile_name = data.get('profile', '')
    accounts = data.get('accounts', [])

    for i, account in enumerate(accounts):
        if account.get('name') == updated.name:
            accounts[i] = updated.to_dict()  # update existing profile
            if updated.is_active and updated.name != active_profile_name:
                data['profile'] = updated.name
            break
    else:
        if updated.is_active:
            data['profile'] = updated.name
        accounts.append(updated.to_dict())  # add new profile

    data['accounts'] = accounts

    with path.open('w') as file:
        yaml.dump(data, file, sort_keys=False)
    return updated


def delete_profile(profile: Union[str, Profile]):
    path = _PROFILE_PATH
    if not path.exists():
        return

    with path.open('r') as file:
        data = yaml.load(file, Loader=yaml.FullLoader)

    accounts = data.get('accounts', [])

    profile_name = profile if isinstance(profile, str) else profile.name
    if profile_name == data.get('profile', ''):
        names = [a.get('name') for a in accounts if a.get('name') != profile_name]
        if len(names) > 0:
            data['profile'] = names[0]  # set next available as active profile
        else:
            path.unlink()  # no more profiles; delete the file
            return

    for i, account in enumerate(accounts):
        if account.get('name') == profile_name:
            accounts.pop(i)
            break

    data['accounts'] = accounts
    with path.open('w') as file:
        yaml.dump(data, file, sort_keys=False)


def get_active_profile() -> Profile:
    profiles = load_profiles()
    for profile in profiles:
        if profile.is_active:
            return profile
    raise NoProfileError()


def _parse_v1(account: dict, active: str) -> 'Profile':
    name = account.get('name', '')
    auth = account.get('auth', {})
    oauth = auth.get('oauth', {})
    return Profile(
        name=name,
        is_active=name == active,
        created_at=account.get('created_at', ''),
        updated_at=account.get('updated_at', ''),
        base_url=account.get('base_url', ''),
        api_key=auth.get('api_key'),
        token=auth.get('token'),
        client_id=oauth.get('client_id'),
        client_secret=oauth.get('client_secret'),
        oauth_path=oauth.get('path'),
        timeout=account.get('timeout'),
        max_retries=account.get('max_retries'),
        retry_interval=account.get('retry_interval'),
        logger=account.get('logger', False),
    )



import json

import click
from cspark.sdk import Client, SparkError
from rich.console import Console

from .._utils import get_active_profile
from ._api import AliasedGroup, header_option, parse_pairs


@click.group(
    name='versions',
    help='Interact with versions of a Spark service',
    options_metavar='',
    cls=AliasedGroup,
    aliases={'ls': 'list', 'get': 'list'},
)
def versions_cmd():
    pass


@versions_cmd.command(name='list', help='List available versions of a service')
@click.argument('uri', type=str, metavar='<SERVICE URI>')
@header_option()
@click.option('--latest', is_flag=True, default=False, help='Show only the latest version')
def list_versions(uri: str, headers: list[str], latest: bool) -> None:
    profile = get_active_profile()
    console = Console()

    try:
        client = Client(**profile.to_config())
        client.config.extra_headers.update(parse_pairs(headers))
        with client.services as s:
            versions = s.get_versions(uri).data

        if latest and isinstance(versions, (list, tuple)) and len(versions) > 0:
            versions = versions[0]

        click.echo(json.dumps(versions, indent=2))
    except SparkError as err:
        console.print(f'[red]✗[/red] {err.message}')
        raise click.exceptions.Exit(1) from err
    except Exception as exc:
        console.print(f'[red]✗[/red] {exc}')
        raise click.exceptions.Exit(1) from exc



from __future__ import annotations

import json
import os
import re
from typing import Mapping, Optional, Union

from ._config import Config
from ._constants import ENV_VARS
from ._errors import SparkError
from ._logger import get_logger
from ._utils import StringUtils
from .resources import AccessToken
from .resources import OAuth2 as OAuthManager

__all__ = ['Authorization', 'OAuth']


class Authorization:
    _oauth: Optional['OAuth']

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        oauth: Union[None, Mapping[str, str], str] = None,
    ) -> None:
        client_id = os.getenv(ENV_VARS.CLIENT_ID)
        client_secret = os.getenv(ENV_VARS.CLIENT_SECRET)
        oauth_path = os.getenv(ENV_VARS.OAUTH_PATH)

        self._api_key = api_key
        self._token = re.sub(r'(?i)\bBearer\b', '', token).strip() if token else None
        if oauth:
            self._oauth = OAuth.when(oauth) if isinstance(oauth, (str, Mapping)) else None
        elif client_id and client_secret:
            self._oauth = OAuth({'client_id': client_id, 'client_secret': client_secret})
        elif oauth_path:
            self._oauth = OAuth.file(oauth_path)
        else:
            self._oauth = None

        if self.is_empty:
            raise SparkError.sdk(
                cause=json.dumps({'api_key': api_key, 'token': token, 'oauth': str(oauth)}),
                message='user authentication is required; '
                'provide a valid API key, bearer token, or OAuth credentials to proceed.\n'
                'If you will be fetching public APIs, set API key as "open".',
            )

    @property
    def is_open(self) -> bool:
        return self._api_key == 'open' or self._token == 'open'

    @property
    def is_empty(self) -> bool:
        return not self._api_key and not self._token and not self._oauth

    @property
    def api_key(self) -> Optional[str]:
        if self._api_key and not self.is_open:
            return StringUtils.mask(self._api_key)
        return self._api_key

    @property
    def token(self) -> Optional[str]:
        return self._token

    @property
    def oauth(self) -> Optional['OAuth']:
        return self._oauth

    @property
    def type(self) -> Optional[str]:
        return 'api_key' if self._api_key else 'token' if self._token else 'oauth' if self._oauth else None

    @property
    def as_header(self) -> Mapping[str, str]:
        if self._oauth:
            return {'Authorization': f'Bearer {self._oauth.access_token}'}
        if self._api_key and not self.is_open:
            return {'x-synthetic-key': self._api_key}
        if self._token and not self.is_open:
            return {'Authorization': f'Bearer {self._token}'}
        return {}


class OAuth:
    _access_token: Optional[AccessToken]

    def __init__(self, value: Mapping[str, str]) -> None:
        self._client_id = value.get('client_id', '')
        self._client_secret = value.get('client_secret', '')
        self._file_path: Optional[str] = value.get('oauth_path')
        self._access_token = None

        if 'client_id' not in value or 'client_secret' not in value:
            raise SparkError.sdk(
                cause=json.dumps(value),
                message='invalid authorization properties. '
                'Provide a JSON object including cliend_id and client_secret '
                'or a path to a JSON file containing the client ID and secret.',
            )

    @staticmethod
    def when(value: Union[Mapping[str, str], str]) -> 'OAuth':
        if isinstance(value, str):
            return OAuth.file(value)
        return OAuth(value)

    @staticmethod
    def file(path: str) -> 'OAuth':
        try:
            with open(path) as file:
                creds = json.load(file)
        except Exception as cause:
            raise SparkError.sdk(f'failed to create oauth credentials from file <{path}>', cause=cause) from cause
        return OAuth({**creds, 'oauth_path': path})

    def __str__(self) -> str:
        return json.dumps(
            {
                'client_id': self._client_id,
                'client_secret': self.client_secret,
                'access_token': self.access_token,
            }
        )

    @property
    def access_token(self) -> Optional[str]:
        return self._access_token.access_token if self._access_token else None

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def client_secret(self) -> str:
        return StringUtils.mask(self._client_secret)

    @property
    def version(self) -> str:
        return '2.0'

    @property
    def flow(self) -> str:
        return 'client_credentials'

    def to_dict(self) -> Mapping[str, str]:
        return {'client_id': self._client_id, 'client_secret': self._client_secret}

    def retrieve_token(self, config: Config) -> AccessToken:
        logger = get_logger(**config.logger.__dict__)

        try:
            with OAuthManager(config) as manager:
                logger.info('retrieving OAuth2 access token...')
                self._access_token = manager.get_access_token()
                if not self._access_token:
                    raise SparkError('no access token found')
                return self._access_token
        except Exception as cause:
            error = SparkError('failed to retrieve OAuth2 access token', cause)
            logger.warning(error.message)
            raise error from cause



import json
import time
from datetime import datetime
from typing import List, Optional, Union, cast

from .._errors import RetryTimeoutError, SparkError
from .._utils import DateUtils, StringUtils, get_retry_timeout, is_int
from ._base import ApiResource, Uri, UriParams

__all__ = ['History']


class History(ApiResource):
    @property
    def downloads(self) -> 'LogDownload':
        return LogDownload(self.config)

    def get(self, call_id: str):
        url = Uri.of(base_url=self.config.base_url.full, endpoint=f'log/getexecutionhistorybycallid/{call_id}')

        response = self.request(url, method='GET')
        if isinstance(response.data, dict):
            formatted = json.loads(response.data.get('response_data', '{}'))
            response = response.copy_with(data=formatted)
        return response

    def rehydrate(
        self,
        uri: Union[None, str, UriParams],
        *,
        call_id: str,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        index: Optional[int] = None,
        legacy: bool = False,
    ):
        if StringUtils.is_empty(call_id):
            raise SparkError.sdk('call_id is required when rehydrating', {'call_id': call_id})

        uri = Uri.validate(UriParams(folder, service) if uri is None else Uri.to_params(uri))
        endpoint = f'download/{call_id}' if legacy else f'download/xml/{call_id}'
        url = Uri.of(uri.pick('folder', 'service'), base_url=self.config.base_url.full, endpoint=endpoint)
        params = {'index': str(index)} if is_int(index) and cast(int, index) >= 0 else None
        response = self.request(url, method='POST', params=params)

        if isinstance(response.data, dict) and isinstance(response.data['response_data'], dict):
            download_url = response.data['response_data']['download_url']
            response.data['status'] = 'Success'  # comes as None from the API
        else:
            raise SparkError('failed to produce a download URL', response)

        return self.request(download_url).copy_with(data=response.data)

    def download(
        self,
        *,
        folder: str,
        service: str,
        type: str = 'json',
        version_id: Optional[str] = None,
        call_ids: Optional[List[str]] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        timezone_offset: Optional[str] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        with self.downloads as downloads:
            response = downloads.initiate(
                folder=folder,
                service=service,
                type=type,
                version_id=version_id,
                call_ids=call_ids,
                start_date=start_date,
                end_date=end_date,
                source_system=source_system,
                correlation_id=correlation_id,
                timezone_offset=timezone_offset,
            )

            job_id = isinstance(response.data, dict) and response.data.get('response_data', {}).get('job_id') or ''
            if not job_id:
                error = SparkError('failed to produce a download job', response)
                self.logger.error(error.message)
                raise error

            job = downloads.get_status(
                job_id=job_id,
                folder=folder,
                service=service,
                type=type,
                max_retries=max_retries,
                retry_interval=retry_interval,
            )

            download_url = isinstance(job.data, dict) and job.data.get('response_data', {}).get('download_url') or ''
            if not download_url:
                error = SparkError(f'failed to produce a download URL for <{job_id}>', job)
                self.logger.error(error.message)
                raise error

            logs = self.request(download_url)

            # NOTE: rewrite response as API response is missing status
            job.data.update({'status': 'Success'})  # type: ignore
            return logs.copy_with(status=job.status, data=job.data)


class LogDownload(ApiResource):
    def initiate(
        self,
        *,
        folder: str,
        service: str,
        type: str = 'json',
        version_id: Optional[str] = None,
        call_ids: Optional[List[str]] = None,
        start_date: Union[None, str, int, datetime] = None,
        end_date: Union[None, str, int, datetime] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        timezone_offset: Optional[str] = None,
    ):
        type = type.lower() if type.lower() in ['json', 'csv'] else 'json'
        uri = Uri.validate(UriParams(folder, service))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=f'log/download{type}')

        call_ids = call_ids or []
        if len(call_ids) == 0 and source_system:
            call_ids.append(source_system)
        if len(call_ids) == 0 and correlation_id:
            call_ids.append(correlation_id)

        body = {
            'request_data': {'call_ids': call_ids, 'timezone_offset': timezone_offset},
            'request_meta': {'version_id': version_id},
        }
        if DateUtils.is_date(start_date):
            body['request_data']['start_date'] = DateUtils.to_datetime(start_date).isoformat()  # type: ignore
        if DateUtils.is_date(end_date):
            body['request_data']['end_date'] = DateUtils.to_datetime(end_date).isoformat()  # type: ignore

        response = self.request(url, method='POST', body=body)
        if isinstance(response.data, dict):
            self.logger.info(f'{type} download job created <{response.data.get("response_data",{}).get("job_id")}>')
        return response

    def get_status(
        self,
        *,
        folder: str,
        service: str,
        job_id: str,
        type: str = 'json',
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        throwable: bool = False,
    ):
        max_retries = max_retries or self.config.max_retries
        retry_interval = retry_interval or self.config.retry_interval
        type = type.lower() if type.lower() in ['json', 'csv'] else 'json'

        uri = Uri.validate(UriParams(folder, service))
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=f'log/download{type}/status/{job_id}')

        retries = 0
        response = self.request(url)
        while True:
            response_data = isinstance(response.data, dict) and response.data.get('response_data', {}) or {}
            progress = response_data.get('progress', 0)

            if progress == 100:
                return response

            if progress < 100 and retries < max_retries:
                self.logger.info(f'waiting for log download status job to complete - {progress}%')

                retries += 1
                time.sleep(get_retry_timeout(retries, retry_interval))
                response = self.request(url)
            else:
                err_msg = f'log download job status check timed out after {retries} attempts'
                if throwable:
                    self.logger.error(err_msg)
                    raise RetryTimeoutError(err_msg, retries=retries, interval=retry_interval)
                self.logger.warning(err_msg)
                return response



import json
import os
import re
from typing import Any, Mapping, Optional, Union, cast
from urllib.parse import urlparse

from httpx import Client as HttpClient

from ._constants import *
from ._errors import SparkError
from ._logger import LoggerOptions
from ._utils import StringUtils
from ._validators import Validators

__all__ = ['Config', 'BaseUrl', 'HealthUrl']


class Config:
    _options: str
    _logger: LoggerOptions

    def __init__(
        self,
        *,
        base_url: Union[None, str, 'BaseUrl'] = None,
        tenant: Optional[str] = None,
        env: Optional[str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        oauth: Union[None, Mapping[str, str], str] = None,
        timeout: Optional[float] = DEFAULT_TIMEOUT_IN_MS,
        max_retries: Optional[int] = DEFAULT_MAX_RETRIES,
        retry_interval: Optional[float] = DEFAULT_RETRY_INTERVAL,
        logger: Union[bool, Mapping[str, Any], LoggerOptions] = True,
        # A custom HTTP client to use with extra capabilities (proxy, auth, verify, etc).
        http_client: Optional[HttpClient] = None,
    ) -> None:
        from ._auth import Authorization  # NOTE: help avoid circular import

        num_validator = Validators.positive_num()

        base_url = os.getenv(ENV_VARS.BASE_URL) if base_url is None else base_url
        api_key = os.getenv(ENV_VARS.API_KEY) if api_key is None else api_key
        token = os.getenv(ENV_VARS.BEARER_TOKEN) if token is None else token

        self._base_url = base_url if isinstance(base_url, BaseUrl) else BaseUrl.of(url=base_url, tenant=tenant, env=env)
        self._auth = Authorization(api_key=api_key, token=token, oauth=oauth)
        self._timeout = timeout if num_validator.is_valid(timeout) else DEFAULT_TIMEOUT_IN_MS
        self._max_retries = max_retries if num_validator.is_valid(max_retries) else DEFAULT_MAX_RETRIES
        self._retry_interval = retry_interval if num_validator.is_valid(retry_interval) else DEFAULT_RETRY_INTERVAL
        self._logger = LoggerOptions.when(logger)

        self.http_client = http_client
        self.extra_headers = {}
        self._options = str(
            {
                'base_url': self._base_url.full,
                'oauth': str(self._auth.oauth) if self._auth.oauth else None,
                'api_key': self._auth.api_key,
                'token': self._auth.token,
                'timeout': self._timeout,
                'max_retries': self._max_retries,
                'retry_interval': self._retry_interval,
                'logger': self._logger,
            }
        )

    def __str__(self) -> str:
        return self._options

    @property
    def has_headers(self) -> bool:
        return len(self.extra_headers) > 0

    @property
    def base_url(self) -> 'BaseUrl':
        return self._base_url

    @property
    def auth(self):
        return self._auth

    @property
    def timeout(self) -> float:
        return cast(float, self._timeout)

    @property
    def max_retries(self) -> int:
        return cast(int, self._max_retries)

    @property
    def retry_interval(self) -> float:
        return cast(float, self._retry_interval)

    @property
    def logger(self) -> LoggerOptions:
        return self._logger

    def copy_with(
        self,
        *,
        base_url: Optional[str] = None,
        tenant: Optional[str] = None,
        env: Optional[str] = None,
        oauth: Union[None, Mapping[str, str], str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[float] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        http_client: Optional[HttpClient] = None,
    ) -> 'Config':
        url = (
            base_url.copy_with(tenant=tenant, env=env)
            if isinstance(base_url, BaseUrl)
            else self.base_url.copy_with(url=base_url, tenant=tenant, env=env)
        )
        return Config(
            base_url=url,
            oauth=oauth or self._auth.oauth.to_dict() if self._auth.oauth else None,
            api_key=api_key or self._auth.api_key,
            token=token or self._auth.token,
            timeout=timeout or self._timeout,
            max_retries=max_retries or self._max_retries,
            retry_interval=retry_interval or self._retry_interval,
            http_client=http_client or self.http_client,
        )

    def get(self):
        """Fetches the SaaS configuration for the current user (via API)."""
        from .resources import Platform

        return Platform(self).get_config()


class BaseUrl:
    __services = ['excel', 'keycloak', 'utility', 'entitystore']
    _service: Optional[str]
    _env: Optional[str]

    def __init__(self, url: str, tenant: str):
        match = re.match(r'https://([^\.]+)\.((?:[^\.]+\.)?[^\.]+)\.coherent\.global', url)
        if match:
            service, env = match.groups()
            self._env = str(env).strip().lower() if env else None
            self._service = str(service).strip().lower() in self.__services and service or 'excel'
            self._base = f'https://{self._service}.{self._env}.coherent.global'
        else:
            self._env = None
            self._service = None
            self._base = url
        self._tenant = tenant

    @property
    def tenant(self) -> str:
        return self._tenant

    @property
    def env(self) -> Optional[str]:
        return self._env

    @property
    def service(self) -> Optional[str]:
        return self._service

    @property
    def full(self) -> str:
        return f'{self._base}/{self._tenant}'

    @property
    def value(self) -> str:
        return self._base

    @property
    def oauth2(self) -> str:
        return f'{self.to("keycloak")}/auth/realms/{self._tenant}'

    def to(self, service: str = 'excel', with_tenant: bool = False) -> str:
        return (self.full if with_tenant else self.value).replace(self._service or 'excel', service)

    def copy_with(
        self, *, url: Optional[str] = None, tenant: Optional[str] = None, env: Optional[str] = None
    ) -> 'BaseUrl':
        tenant, env = tenant or self._tenant, env or self._env
        return BaseUrl.of(url=url, tenant=tenant) if url else BaseUrl.of(tenant=tenant, env=env)

    @staticmethod
    def of(*, url: Optional[str] = None, tenant: Optional[str] = None, env: Optional[str] = None) -> 'BaseUrl':
        str_validator = Validators.empty_str()
        url_validator = Validators.base_url()

        if url_validator.is_valid(url):
            parsed_url = urlparse(url.rstrip('/'))  # pyright: ignore[reportOptionalMemberAccess]
            paths = str(parsed_url.path).split('/')
            maybe_tenant = paths[1] if len(paths) > 1 else tenant

            if str_validator.is_valid(maybe_tenant, 'tenant name is required'):
                base_url = f'{parsed_url.scheme}://{parsed_url.netloc}'
                return BaseUrl(url=base_url, tenant=str(maybe_tenant))
        elif StringUtils.is_not_empty(tenant) and StringUtils.is_not_empty(env):
            base_url = f'https://excel.{str(env).strip().lower()}.coherent.global'
            return BaseUrl(url=base_url, tenant=str(tenant).strip().lower())
        else:
            # capture errors for missing parameters
            str_validator.is_valid(env, 'environment name is missing') and str_validator.is_valid(
                tenant, 'tenant name is missing'
            )  # pyright: ignore[reportUnusedExpression]

        errors = url_validator.errors  # + str_validator.errors
        raise SparkError.sdk(
            message='; '.join(e.message for e in errors)
            if len(errors) > 0
            else 'cannot build base URL from invalid parameters',
            cause=json.dumps({'url': url, 'tenant': tenant, 'env': env}),
        )


class HealthUrl(BaseUrl):
    def __init__(self, base_url: str):
        super().__init__(base_url, tenant='')

    @staticmethod
    def when(url: Union[str, BaseUrl]) -> 'HealthUrl':
        parsed_url = urlparse(url if isinstance(url, str) else url.to('excel'))
        if parsed_url.scheme:
            return HealthUrl(f'{parsed_url.scheme}://{parsed_url.netloc}')
        else:
            return HealthUrl(f'https://excel.{url}.coherent.global')  # otherwise treat as environment
